from enum import Enum

from sklearn.decomposition import PCA
from sklearn.manifold import TSNE, Isomap
from umap import UMAP


class ReferenceGroup(Enum):
    USER_DEFINED = 1
    NOVELTY_DETECTION = 2


class TimeUnit(Enum):
    HOUR = 1/24
    DAY = 1
    MONTH = 30
    YEAR = 365


class FeatureColumnsType(Enum):
    BACTERIA = 1
    METADATA = 2
    BACTERIA_AND_METADATA = 3


class AnomalyType(Enum):
    PREDICTION_INTERVAL = 1
    LOW_PASS_FILTER = 2
    ISOLATION_FOREST = 3


class FeatureExtraction(Enum):
    NONE = 0
    NEAR_ZERO_VARIANCE = 1
    CORRELATION = 2
    TOP_K_IMPORTANT = 3


class Normalization(Enum):
    NORMALIZED = True
    NON_NORMALIZED = False


class EmbeddingModelType(Enum):
    PCA = PCA
    TSNE = TSNE
    UMAP = UMAP
    ISOMAP = Isomap



import copy
import itertools
from collections import defaultdict
from itertools import combinations

import numpy as np
import pandas as pd
import plotly.express as px
import plotly.graph_objects as go
import scipy as sp
import scipy.stats as stats
import shap
import math
from natsort import natsorted
from plotly.subplots import make_subplots
from sklearn.ensemble import IsolationForest, RandomForestRegressor
from sklearn.feature_selection import VarianceThreshold
from sklearn.metrics import mean_absolute_error, r2_score
from sklearn.model_selection import (
    GridSearchCV,
    GroupKFold,
    GroupShuffleSplit,
    cross_val_score,
)
from sklearn.preprocessing import StandardScaler

from microbiome.enumerations import (
    AnomalyType,
    FeatureColumnsType,
    FeatureExtraction,
    TimeUnit,
)
from microbiome.statistical_analysis import permuspliner, regliner

RANDOM_STATE = 42


class MicrobiomeTrajectory:
    def __init__(
        self,
        dataset,
        feature_columns,
        feature_extraction=FeatureExtraction.NONE,
        time_unit=TimeUnit.DAY,
        anomaly_type=AnomalyType.PREDICTION_INTERVAL,
        train_indices=None,
    ):
        self.dataset = copy.deepcopy(dataset)
        # initialize feature columns (only once, in constructor!)
        if isinstance(feature_columns, (list, np.ndarray)):
            self._feature_columns = feature_columns
        elif feature_columns == FeatureColumnsType.BACTERIA:
            self._feature_columns = self.dataset.bacteria_columns
        elif feature_columns == FeatureColumnsType.METADATA:
            self._feature_columns = self.dataset.metadata_columns
        elif feature_columns == FeatureColumnsType.BACTERIA_AND_METADATA:
            self._feature_columns = self.dataset.bacteria_and_metadata_columns

        self.axis_settings_default = dict(
            tick0=0,
            mirror=True,
            # dtick=2,
            showline=True,
            linecolor="lightgrey",
            gridcolor="lightgrey",
            zeroline=True,
            zerolinecolor="lightgrey",
            showspikes=True,
            spikecolor="gray",
            autorange=True,
        )

        results = self.get_less_feature_columns(plot=True, technique=feature_extraction)
        self.feature_columns = results["feature_columns"]
        self.feature_importance = results["feature_importance"]
        self.feature_columns_plot = results["fig"]
        self.feature_columns_plot_ret_val = results["ret_val"]
        self.feature_columns_plot_config = results["config"]

        self.dataset.time_unit = time_unit
        self.anomaly_type = anomaly_type

        # TODO: handle less features here too!
        if train_indices is None:
            train_indices = (self.dataset.df.reference_group == True).values

        df_train = self.dataset.df.iloc[train_indices]
        df_test = self.dataset.df.iloc[~train_indices]

        X_train = df_train[self.feature_columns].values
        y_train = df_train.age_at_collection.values

        groups_train = list(df_train.subjectID.values)

        # estimator = RandomForestRegressor(random_state=RANDOM_STATE)
        # estimator.fit(X, y)
        parameters_gridsearch = {"n_estimators": [50, 100, 150]}
        rfr = RandomForestRegressor(random_state=RANDOM_STATE)
        n_splits = 5 if len(np.unique(groups_train)) > 5 else len(np.unique(groups_train))
        gkf = list(GroupKFold(n_splits=n_splits).split(X_train, y_train, groups=groups_train))
        search = GridSearchCV(rfr, parameters_gridsearch, cv=gkf)
        search.fit(X_train, y_train)
        self.estimator = search.best_estimator_

        y_train_pred = self.estimator.predict(X_train).astype(float)
        self.dataset.df.loc[df_train.index, "MMI"] = y_train_pred

        if len(df_test) > 0:
            X_test = df_test[self.feature_columns].values
            y_test = df_test.age_at_collection.values
            y_test_pred = self.estimator.predict(X_test).astype(float)
            self.dataset.df.loc[df_test.index, "MMI"] = y_test_pred

        self.reference_groups = self.dataset.df.reference_group.values.astype(bool)
        self.X = self.dataset.df[self.feature_columns]
        self.y = self.dataset.df.age_at_collection.values.astype(float)
        self.y_pred = self.dataset.df.MMI.values.astype(float)
        self.sample_ids = self.dataset.df.sampleID.values.astype(str)
        self.subject_ids = self.dataset.df.subjectID.values.astype(str)
        self.groups = self.dataset.df.group.values.astype(str)
        self.x = np.linspace(np.min(self.y), np.max(self.y), 100)

        self.layout_settings_default = dict(
            height=900,
            width=1200,
            plot_bgcolor="rgba(255,255,255,255)",
            paper_bgcolor="rgba(255,255,255,255)",
            margin=dict(l=70, r=70, t=70, b=70),
            font=dict(size=17),
            hoverdistance=-1,
            legend=dict(
                x=1.01,
                y=1,
            ),
        )

        self.color_reference = "26,150,65"
        self.color_non_reference = "255,150,65"
        self.color_anomaly = "255,0,0"
        colors = px.colors.qualitative.Plotly
        colors_rgb = [
            tuple(int(h.lstrip("#")[i : i + 2], 16) for i in (0, 2, 4)) for h in colors
        ]
        colors_rgb = [str(x)[1:-1] for x in colors_rgb]
        self.colors_rgb = colors_rgb

        self.bacteria_colors = {}
        self.group_colors = {}
        self.subject_colors = {}
        self.palette = itertools.cycle(self.colors_rgb)

    def get_less_feature_columns(
        self,
        plot=False,
        technique=FeatureExtraction.NONE,
        thresholds=None,
        layout_settings=None,
        xaxis_settings=None,
        yaxis_settings=None,
    ):
        """Extract just few feature columns to reduce the number of features wlog.

        Parameters
        ----------
        plot : bool, optional
            Whether to plot the performance statistics, by default False.
            If True, the plot with MAEs and R2s w.r.t. the number of features used is returned.
        technique : FeatureExtraction, str, optional
            Technique used to reduce number of features, by default FeatureExtraction.NONE.
            Possible values: NONE, NEAR_ZERO_VARIANCE, CORRELATION, TOP_K_IMPORTANT
        thresholds : list, optional
            Threshold values to be tested for performance, by default None.
        layout_settings : dict, optional
            Layout settings fot the plotly, by default None.

        Returns
        -------
        results : dict
            Dictionary with the following keys:
            - feature_columns: list of feature columns found used 3 reduction techniques.
            - fig : figure object of performances with different number of feature columns.
            - config : plotly config object.
        """
        fig = None
        config = None
        ret_val = None

        if technique == FeatureExtraction.NONE:
            feature_columns = self._feature_columns
        else:
            X = self.dataset.df[self._feature_columns].values
            y = self.dataset.df.age_at_collection.values
            groups = self.dataset.df.subjectID.values

            estimator = RandomForestRegressor(random_state=RANDOM_STATE)
            estimator.fit(X, y)

            scorer_list = ["r2", "neg_mean_squared_error"]
            scores = defaultdict(list)

            # get thresholds
            if thresholds is None:
                if technique == FeatureExtraction.TOP_K_IMPORTANT:
                    feature_importances = np.array(estimator.feature_importances_)
                    feature_importance = pd.DataFrame(
                        list(zip(self._feature_columns, feature_importances)),
                        columns=["feature_name", "feature_importance"],
                    )
                    feature_importance.sort_values(
                        by=["feature_importance"], ascending=False, inplace=True
                    )
                    important_features = feature_importance.feature_name.values
                    thresholds = np.linspace(1, 50, num=10, dtype=int)
                elif technique == FeatureExtraction.NEAR_ZERO_VARIANCE:
                    variances_ = np.nanvar(X, axis=0)
                    thresholds = np.linspace(
                        0, np.max(variances_), num=10, endpoint=False
                    )
                elif technique == FeatureExtraction.CORRELATION:
                    thresholds = np.linspace(0, 1, num=10, endpoint=False)

            estimator = RandomForestRegressor(random_state=RANDOM_STATE)
            for threshold in thresholds:
                if technique == FeatureExtraction.TOP_K_IMPORTANT:
                    feature_columns = important_features[:threshold]
                elif technique == FeatureExtraction.NEAR_ZERO_VARIANCE:
                    X = self.dataset.df[self._feature_columns].values
                    constant_filter = VarianceThreshold(threshold=threshold)
                    constant_filter.fit(X)
                    idx = np.where(constant_filter.get_support())[0]
                    feature_columns = self._feature_columns[idx]
                elif technique == FeatureExtraction.CORRELATION:
                    correlation_matrix = (
                        self.dataset.df[self._feature_columns].corr().abs()
                    )
                    upper = correlation_matrix.where(
                        np.triu(np.ones(correlation_matrix.shape), k=1).astype(np.bool)
                    )
                    correlated_features = [
                        column
                        for column in upper.columns
                        if any(upper[column] > threshold)
                    ]
                    idx = np.isin(
                        self._feature_columns, correlated_features, invert=True
                    )
                    feature_columns = self._feature_columns[idx]

                X = self.dataset.df[feature_columns].values
                for scorer in scorer_list:
                    cv_scores = cross_val_score(
                        estimator,
                        X,
                        y,
                        scoring=scorer,
                        groups=groups,
                        cv=GroupShuffleSplit(random_state=RANDOM_STATE),
                    )
                    scores["score"].extend(cv_scores)
                    scores["scorer"].extend([scorer] * len(cv_scores))
                    scores["features_number"].extend(
                        [len(feature_columns)] * len(cv_scores)
                    )

                    scores["threshold"].extend([threshold] * len(cv_scores))

            df_scores = pd.DataFrame(data=scores)
            r2_mean = (
                df_scores[df_scores.scorer == "r2"]
                .groupby(by="features_number")
                .mean()["score"]
            )
            i = np.argmax(r2_mean.values)

            # excract only top important features
            thresholds = (
                df_scores[df_scores.scorer == "r2"]
                .groupby(by="features_number")
                .first()["threshold"]
                .values
            )
            threshold = thresholds[i]

            # get feature_columns
            if technique == FeatureExtraction.TOP_K_IMPORTANT:
                feature_columns = important_features[:threshold]
            elif technique == FeatureExtraction.NEAR_ZERO_VARIANCE:
                X = self.dataset.df[self._feature_columns].values
                constant_filter = VarianceThreshold(threshold=threshold)
                constant_filter.fit(X)
                idx = np.where(constant_filter.get_support())[0]
                feature_columns = self._feature_columns[idx]
            elif technique == FeatureExtraction.CORRELATION:
                correlation_matrix = self.dataset.df[self._feature_columns].corr().abs()
                upper = correlation_matrix.where(
                    np.triu(np.ones(correlation_matrix.shape), k=1).astype(np.bool)
                )
                correlated_features = [
                    column for column in upper.columns if any(upper[column] > threshold)
                ]
                idx = np.isin(self._feature_columns, correlated_features, invert=True)
                feature_columns = self._feature_columns[idx]

            if plot:
                layout_settings_default = dict(
                    height=500,
                    width=400 * len(scorer_list),
                    plot_bgcolor="rgba(255,255,255,255)",
                    paper_bgcolor="rgba(255,255,255,255)",
                    margin=dict(l=0, r=0, b=0, pad=0),
                    title_text=f"{technique.name.title()} features",
                    font=dict(size=12),
                    hovermode="x",
                )
                if layout_settings is None:
                    layout_settings = {}
                layout_settings_final = {**layout_settings_default, **layout_settings}
                if xaxis_settings is None:
                    xaxis_settings = {}
                if yaxis_settings is None:
                    yaxis_settings = {}
                xaxis_settings_final = {**self.axis_settings_default, **xaxis_settings}
                yaxis_settings_final = {**self.axis_settings_default, **yaxis_settings}

                fig = make_subplots(
                    rows=1, cols=len(scorer_list), horizontal_spacing=0.2
                )
                ret_val = "<b>Performance Information</b><br>"

                for col, scorer in enumerate(scorer_list, start=1):
                    df_scorer = df_scores[df_scores.scorer == scorer]
                    scorer_mean = df_scorer.groupby(by="features_number").mean()[
                        "score"
                    ]
                    scorer_std = df_scorer.groupby(by="features_number").std()["score"]
                    thresholds = df_scorer.groupby(by="features_number").first()[
                        "threshold"
                    ]
                    hovertemplate = (
                        "Features number: %{x}<br>R2 Score: %{y}<br>"
                        + "Threshold: %{text}"
                    )
                    fig.add_trace(
                        go.Scatter(
                            x=scorer_mean.index,
                            y=scorer_mean.values,
                            error_y=dict(
                                type="data",
                                array=scorer_std.values,
                                visible=True,
                            ),
                            name=scorer,
                            hovertemplate=hovertemplate,
                            text=thresholds,
                        ),
                        row=1,
                        col=col,
                    )

                    rangey = (0, 1) if scorer == "r2" else None
                    showlegend = scorer == "r2"

                    ret_val += (
                        f"<b>{scorer.title()}</b><br>"
                        + f"Performance: {scorer_mean.values[i]:.3f}±{scorer_std.values[i]:.3f}<br>"
                        + f"Features number: {scorer_mean.index[i]}<br>"
                    )

                    fig.add_trace(
                        go.Scatter(
                            x=[scorer_mean.index[i]],
                            y=[scorer_mean.values[i]],
                            mode="markers",
                            marker_symbol="star",
                            marker_size=15,
                            marker_color="green",
                            name="optimal r2",
                            showlegend=showlegend,
                            hovertemplate="%{y:.0%}",
                        ),
                        row=1,
                        col=col,
                    )
                    fig.update_xaxes(
                        title="Number of features",
                        row=1,
                        col=col,
                        **xaxis_settings_final,
                    )
                    fig.update_yaxes(
                        title=scorer,
                        row=1,
                        col=col,
                        # range=rangey,
                        **yaxis_settings_final,
                    )

                fig.update_layout(**layout_settings_final)
                config = {
                    "toImageButtonOptions": {
                        "format": "svg",  # one of png, svg, jpeg, webp
                        "filename": "bacteria_abundances",
                        "height": layout_settings_final["height"],
                        "width": layout_settings_final["width"],
                        "scale": 1,  # Multiply title/legend/axis/canvas sizes by this factor
                    }
                }

        X = self.dataset.df[feature_columns].values
        y = self.dataset.df.age_at_collection.values
        groups = self.dataset.df.subjectID.values

        estimator = RandomForestRegressor(random_state=RANDOM_STATE)
        estimator.fit(X, y)

        feature_importances = np.array(estimator.feature_importances_)
        feature_importance = pd.DataFrame(
            list(zip(feature_columns, feature_importances)),
            columns=["feature_name", "feature_importance"],
        )
        feature_importance.sort_values(
            by=["feature_importance"], ascending=False, inplace=True
        )

        results = {
            "fig": fig,
            "config": config,
            "ret_val": ret_val,
            "feature_columns": list(feature_columns),
            "feature_importance": feature_importance,
        }

        return results

    def get_prediction_and_confidence_interval(self, indices, x=None, degree=2):
        """Get prediction and confidence interval for samples indicated by indices.

        Parameters
        ----------
        indices : list
            Indices of samples that will be used to get the prediction and confidence interval.
        x : numpy.array
            Different sampling points for the age_at_collection values. By default it is None and
            the give age_at_collection values are used (vector y).
        degree : int
            Degree of the polynomial used to fit the data. By default it is 2.

        Returns
        -------
        results : dict
            Dictionary with the following keys:
                - 'prediction_interval' : dict
                    Prediction interval for each sample sampled at x given by the mean and +- bounds.
                - 'confidence_interval' : dict
                    Confidence interval for each sample sampled at x given by the mean and +- bounds.

        Attributes
        ----------
        q : float
            Confidence level for the confidence interval. By default it is 0.975.
        p : numpy.array
            Polynomial coefficients, highest power first.
        y_model : numpy.array
            Model of the polynomial fit to the data.
        n : int
            Number of samples (observations).
        m : int
            Number of features (parameters).
        dof : int
            Degrees of freedom.
        t : float
            Student's t value. Used for CI and PI bands.
        residual : numpy.array
            Residuals of the polynomial fit. Estimates the error in the data.
        s_err : numpy.array
            Standard deviation of the error.
        y2 : numpy.array
            Prediction interval mean.
        pi : numpy.array
            Prediction interval bound.
        n_boot : int
            Number of bootstrap samples.
        y3 : numpy.array
            Confidence interval mean of every bootstraped version.
        """
        if x is None:
            x = copy.deepcopy(self.y)
            x = x[indices]
        y = self.y[indices]
        y_pred = self.y_pred[indices]

        # Stats
        p = np.polyfit(y, y_pred, degree)
        y_model = np.polyval(p, y)

        # Statistics
        q = 0.975
        n = y_pred.size
        m = p.size
        dof = n - m
        t = stats.t.ppf(q, n - m)
        residual = y_pred - y_model
        # chi2 = np.sum((residual / y_model)**2)  # chi-squared; estimates error in data
        # chi2_red = chi2 / dof                   # reduced chi-squared; measures goodness of fit
        s_err = np.sqrt(np.sum(residual ** 2) / dof)

        # Prediction Interval
        y2 = np.polyval(p, x)
        pi = (
            t
            * s_err
            * np.sqrt(1 + 1 / n + (x - np.mean(y)) ** 2 / np.sum((y - np.mean(y)) ** 2))
        )

        # Confidence Interval
        n_boot = 500
        bootindex = sp.random.randint
        y3 = np.empty((n_boot, len(x)))
        for b in range(n_boot):
            resamp_residual = residual[bootindex(0, len(residual) - 1, len(residual))]
            # make coefficients of for polys
            pc = np.polyfit(y, y_pred + resamp_residual, degree)
            # collect bootstrap cluster
            y3[b] = np.polyval(pc, x)

        results = {
            "prediction_interval": {
                "x": x,
                "mean": y2,
                "bound": pi,
            },
            "confidence_interval": {
                "x": x,
                "mean": np.mean(y3, axis=0),
                "bound": 1.96 * np.std(y3, axis=0),
            },
        }
        return results

    def get_anomalies(self, anomaly_type, indices):
        """Get anomalies for samples indicated by indices.

        Parameters
        ----------
        anomaly_type : str
            Type of anomaly to be detected.
        indices : list
            Indices of samples that will be used to identify anomalous samples among them.
            E.g. indices of all healthy samples where we are interested to find the anomalies.

        Returns
        -------
        anomalies : numpy.array
            Indices of the anomalous samples. Indices are relative to the original data.
            Possible values are: True - anomaly, False - non-anomaly, None - other samples that were
            not indicated with indices argument.

        Notes
        -----
        We use three different types to detect anomalies:
            1. Prediction interval
            2. Low-pass filter
            3. Isolation forest
        """
        y_pred = self.y_pred[indices]

        if anomaly_type == AnomalyType.PREDICTION_INTERVAL:
            degree = 3

            results = self.get_prediction_and_confidence_interval(
                indices=indices, degree=degree
            )

            # pi_x = results["prediction_interval"]["x"] == y
            pi_mean = results["prediction_interval"]["mean"]
            pi_bound = results["prediction_interval"]["bound"]

            # highlight_outliers
            anomaly = np.logical_or(
                np.less(y_pred, pi_mean - pi_bound),
                np.greater(y_pred, pi_mean + pi_bound),
            )

        elif anomaly_type == AnomalyType.LOW_PASS_FILTER:
            window = 10
            number_of_stds = 2

            y_pred_rolling_avg = (
                pd.Series(y_pred)
                .rolling(window=window, min_periods=1, center=True)
                .mean()
            )
            y_pred_rolling_std = (
                pd.Series(y_pred)
                .rolling(window=window, min_periods=1, center=True)
                .std(skipna=True)
            )
            anomaly = abs(y_pred - y_pred_rolling_avg) > (
                number_of_stds * y_pred_rolling_std
            )
        elif anomaly_type == AnomalyType.ISOLATION_FOREST:
            outliers_fraction = 0.1
            window = 5

            y_pred_rolling_avg = (
                pd.Series(y_pred)
                .rolling(window=window, min_periods=1, center=True)
                .mean()
            )
            y_pred_zscore = y_pred - y_pred_rolling_avg

            # Subset the dataframe by desired columns
            dataframe_filtered_columns = pd.DataFrame(
                data={
                    "y_pred": y_pred,
                    "y_pred_zscore": y_pred_zscore,
                }
            )

            # Scale the column that we want to flag for anomalies
            np_scaled = StandardScaler().fit_transform(dataframe_filtered_columns)
            scaled_time_series = pd.DataFrame(np_scaled)

            # train isolation forest
            model = IsolationForest(contamination=outliers_fraction)
            model.fit(scaled_time_series)

            # generate column for Isolation Forest-detected anomalies
            anomaly = model.predict(scaled_time_series)
            anomaly[anomaly == 1] = False  # inliner
            anomaly[anomaly == -1] = True  # outliner
            anomaly = anomaly.astype(bool)

        anomalies = np.empty(len(self.y_pred))
        anomalies[indices] = anomaly

        return anomalies

    def get_pvalue_linear(self, column, indices):
        """Get p-value for linear lines.

        Parameters
        ----------
        indices : list
            Indices of samples that contain only 2 groups that will be compared.

        Returns
        -------
        result["k"] : float
            P-value for linear line slopes.
        result["n"] : float
            P-value for linear line y-intercepts.

        Notes
        -----
        The cutoff for significance used is an alpha of 0.05.
        If the p-value is less than 0.05, we reject the null hypothesis that
        there's no difference between the means and conclude that a significant
        difference does exist.
        Below 0.05, significant. Over 0.05, not significant.
        """
        y = self.y[indices]
        y_pred = self.y_pred[indices]
        groups = getattr(self, column)[indices]

        group_values = np.unique(groups)
        print("group_values", group_values)

        assert len(group_values) == 2, f"Needs to have only 2 unique groups to compare, has {group_values}"

        df_stats = pd.DataFrame(
            data={
                "Input": list(y),
                "Output": list(y_pred),
                "Condition": list(groups),
            }
        )

        result = regliner(df_stats, {group_values[0]: 0, group_values[1]: 1})
        return result["k"], result["n"]

    def get_pvalue_spline(self, column, indices, degree=2):
        """Get p-value for spline lines.

        Parameters
        ----------
        indices : list
            Indices of samples that contain only 2 groups that will be compared.
        degree : int
            The degree of the polynomial.

        Returns
        -------
        result["pval"] : float
            P-value for spline lines.

        Notes
        -----
        The cutoff for significance used is an alpha of 0.05.
        If the p-value is less than 0.05, we reject the null hypothesis that
        there's no difference between the means and conclude that a significant
        difference does exist.
        Below 0.05, significant. Over 0.05, not significant.
        """
        y = self.y[indices]
        y_pred = self.y_pred[indices]
        groups = getattr(self, column)[indices]
        sample_ids = self.sample_ids[indices]

        group_values = np.unique(groups)

        assert len(group_values) == 2, f"Needs to have only 2 unique groups to compare, has {group_values}"

        df_stats = pd.DataFrame(
            data={
                "Input": list(y),
                "Output": list(y_pred),
                "Condition": list(groups),
                "sampleID": list(sample_ids),
            }
        )

        result = permuspliner(
            df_stats,
            xvar="Input",
            yvar="Output",
            category="Condition",
            degree=degree,
            cases="sampleID",
            groups=group_values,
            perms=500,
            test_direction="more",
            ints=1000,
            quiet=True,
        )

        return result["pval"]

    def get_top_bacteria_in_time(self, indices, time_current, time_delta):
        """Get top important bacteria for indicated time block.

        Time block is the time interval representing the different bacteria importance
        with their corresponding average and stds.

        Parameters
        ----------
        indices : list
            Indices of samples that will be used to calculate the time block
            averages. E.g. healthy samples without anomalies.
        time_current : float
            Current time point. Time point from which to collect samples into
            time block.
        time_delta : float
            Time delta from current time point. Time delta during which to collect.

        Returns
        -------
        results : dict
            Dictionary with the following keys:
            - "feature_importance" : pandas.DataFrame
                DataFrame with bacteria in the rows, and their importance, average and std in the columns.
            - "df_time_block" : pandas.DataFrame
                DataFrame with all data used to calculate the time block bacterias importance and averages.
        """
        indices_timebox = (time_current <= self.y) & (
            self.y < time_current + time_delta
        )
        i = 0
        while len(indices_timebox) == 0:
            indices_timebox = (time_current - (2 * i - 1) * time_delta <= self.y) & (
                self.y < time_current + (1 + i) * time_delta
            )
            i += 0.1

        indices = np.logical_and(indices_timebox, indices)

        df_time_block = self.dataset.df.iloc[indices][
            self.feature_columns + ["age_at_collection", "MMI"]
        ]

        # average, std, of samples in the time block for a given bacteria
        avgs = df_time_block[self.feature_columns].mean(axis=0)
        stds = df_time_block[self.feature_columns].std(axis=0)
        shap_values = shap.TreeExplainer(self.estimator).shap_values(
            pd.DataFrame([avgs])
        )
        feature_importance = pd.DataFrame(
            data={
                "importance": np.abs(shap_values).mean(0),
                "feature_avg": avgs,
                "feature_std": stds,
            }
        )
        feature_importance.sort_values(by=["importance"], ascending=True, inplace=True)

        results = {
            "feature_importance": feature_importance,
            "df_time_block": df_time_block,
        }

        return results

    def get_bacteria_colors(self, val):
        if not self.bacteria_colors.get(val, None):
            self.bacteria_colors[val] = next(self.palette)
        return self.bacteria_colors[val]

    def get_subject_colors(self, val):
        if not self.subject_colors.get(val, None):
            self.subject_colors[val] = next(self.palette)
        return self.subject_colors[val]

    def get_group_colors(self, val):
        if not self.group_colors.get(val, None):
            self.group_colors[val] = next(self.palette)
        return self.group_colors[val]

    def get_intervention_point(self, X_i, y_i, time_block_ranges, indices):
        """Get intervention point.

        X_i : numpy.array
            Features of i-th sample.
        y_i : float
            Age at collection of i-th sample.
        time_block_ranges : list
            Units or time block sizes.
        indices : list
            Indices of samples that will be used to calculate the time block
            averages. E.g. healthy samples without anomalies.
        """
        normal_vals = [False]

        current_time = time_block_ranges[0]
        for time in time_block_ranges[1:]:
            time_delta = time - current_time
            if current_time <= y_i < current_time + time_delta:
                break

            current_time += time_delta

        shap_values = shap.TreeExplainer(self.estimator).shap_values(
            pd.DataFrame([X_i])
        )

        feature_importance_outlier = pd.DataFrame(
            data={
                "importance_outlier": np.abs(shap_values).mean(0),
                "outlier": X_i,
            },
            index=self.feature_columns,
        )
        feature_importance_outlier.sort_values(
            by=["importance_outlier"], ascending=True, inplace=True
        )

        results = self.get_top_bacteria_in_time(
            indices, time_current=current_time, time_delta=time_delta
        )
        feature_importance = results["feature_importance"]

        feature_importance = feature_importance.join(feature_importance_outlier)
        feature_importance["normal"] = feature_importance.apply(
            lambda x: True
            if x["feature_avg"] - x["feature_std"]
            < x["outlier"]
            < x["feature_avg"] + x["feature_std"]
            else False,
            axis=1,
        )
        feature_importance.sort_values(
            by=["importance_outlier", "importance"], ascending=False, inplace=True
        )
        feature_importance["change"] = feature_importance.apply(
            lambda row: row["feature_avg"] if row["normal"] in normal_vals else None,
            axis=1,
        )
        feature_importance = feature_importance[feature_importance["change"].notna()]

        return feature_importance

    def add_timeboxes(self, fig, time_block_ranges, indices, num_top_bacteria=5):
        """Add time boxes to the figure.

        Parameters
        ----------
        fig : plotly.graph_objects.Figure
            Figure to which the time boxes will be added.
        time_block_ranges : list
            Units or time block sizes.
        indices : list
            Indices of samples that will be used to calculate the time block
            averages. E.g. healthy samples without anomalies.
        num_top_bacteria : int
            Number of top bacteria to be added to the figure per time block.

        Returns
        -------
        fig : plotly.graph_objects.Figure
            Figure with added time boxes.
        """
        current_time = time_block_ranges[0]
        box_height = self._get_axis_max_limit(fig, "y") // 3

        legend_labels = []
        first_iteration = True
        for time in time_block_ranges[1:]:
            time_delta = time - current_time
            box_width = 0.9 * time_delta
            results = self.get_top_bacteria_in_time(indices, current_time, time_delta)
            feature_importance = results["feature_importance"]
            df_time_block = results["df_time_block"]
            x = current_time + time_delta / 2
            y = df_time_block["MMI"].median()
            number_of_samples = len(df_time_block)

            if number_of_samples > 0:

                ratios = feature_importance["importance"].values[-num_top_bacteria:]
                bacteria_names = feature_importance.index.values[-num_top_bacteria:]
                ratios /= sum(ratios)
                if x is not None:
                    y -= box_height / 2
                    x -= box_width / 2
                    y = max(y, 0)
                    # x = max(x, 0)

                    for j in range(num_top_bacteria):
                        bacteria_name = bacteria_names[j]
                        # text mean and std for this bacteria inside one boxplot
                        bacteria_avg = feature_importance["feature_avg"][bacteria_name]
                        bacteria_std = feature_importance["feature_std"][bacteria_name]

                        height = ratios[j] * box_height
                        if bacteria_name not in legend_labels:
                            legend_labels.append(bacteria_name)
                            showlegend = True
                        else:
                            showlegend = False

                        bar_trace = dict(
                            x=[x],
                            y=[height],
                            base=y,
                            offset=0,
                            width=box_width,
                            marker_color=f"rgba({self.get_bacteria_colors(bacteria_name)},0.8)",
                            legendgroup=bacteria_name,
                            showlegend=showlegend,
                            name=self.dataset.nice_name(bacteria_name),
                            hovertemplate="<br>".join(
                                [
                                    f"<b>bacteria: {self.dataset.nice_name(bacteria_name)}</b>",
                                    f"avg ± std: {bacteria_avg:.5f} ± {bacteria_std:.5f}",
                                    f"importance: {ratios[j]*100:.2f}%",
                                    f"# samples: {number_of_samples}",
                                ]
                            ),
                        )
                        if first_iteration:
                            bar_trace = {
                                "legendgroup": "<b>Important bacteria</b>",
                                "legendgrouptitle_text": "<b>Important bacteria</b>",
                                **bar_trace,
                            }
                            first_iteration = False
                        fig.add_trace(go.Bar(bar_trace))

                        y += height

            current_time += time_delta

        return fig

    def add_intervention(self, fig, time_block_ranges, indices):
        """Add intervention to the figure.

        Parameters
        ----------
        fig : plotly.graph_objects.Figure
            Figure to which the intervention will be added.
        time_block_ranges : list
            Units or time block sizes.
        indices : list
            Indices of samples that will be used to calculate the time block
            averages. E.g. healthy samples without anomalies.

        Returns
        -------
        fig : plotly.graph_objects.Figure
            Figure with added intervention.
        """
        number_of_changes = 5

        # click anomaly to intervene
        # our custom event handler
        def _update_trace(trace, points, selector):
            results = self.selection_update_trace(
                fig, time_block_ranges, number_of_changes, indices
            )(trace, points, selector)

        # we need to add the on_click event to each trace separately
        for i in range(len(fig.data)):
            fig.data[i].on_click(_update_trace)

        return fig

    def selection_update_trace(
        self, fig, time_block_ranges, number_of_changes, indices
    ):
        # this list stores the points which were clicked on
        # in all but one trace they are empty
        # if len(points.point_inds) == 0:
        #     return
        if isinstance(fig, dict):
            fig = go.FigureWidget(fig)

        # traces = len(fig.data)

        def _inner(trace, points, selector):
            ret_val = ""
            config = None

            if trace["name"] == "Samples":
                try:
                    if hasattr(points, "point_inds"):
                        idx = points.point_inds[0]
                    else:
                        idx = points
                except:
                    pass

                y = trace["x"]
                y_pred = trace["y"]
                subject_ids = [x[0] for x in list(trace["customdata"])]
                sample_ids = [x[1] for x in list(trace["customdata"])]
                X = pd.DataFrame(
                    [x[2] for x in list(trace["customdata"])],
                    columns=self.feature_columns,
                )

                X_i = X.iloc[idx]
                y_i = y[idx]

                feature_importance = self.get_intervention_point(
                    X_i,
                    y_i,
                    time_block_ranges,
                    indices,
                )

                ret_val = "<b>Intervention bacteria "
                ret_val += f"for Subject {subject_ids[idx]}, Sample {sample_ids[idx]}</b><br><br>"
                ret_val += "<br>".join(
                    [
                        f"<b>{self.dataset.nice_name(x[0])}</b>: {x[1]:.5f} → {x[2]:.5f}"
                        for x in feature_importance[["outlier", "change"]]
                        .reset_index()
                        .values[:number_of_changes]
                    ]
                )

                change = feature_importance.iloc[:number_of_changes]["change"].to_dict()
                for feature in change:
                    X_i[feature] = change[feature]
                X_i = X_i.values.reshape(1, -1)
                sample_pred_new = self.estimator.predict(X_i)

                sample_ids_new = np.append(sample_ids, sample_ids[idx])
                if len(sample_ids_new) == 1 + len(set(sample_ids_new)):
                    y_new = np.append(y, y[idx])
                    y_pred_new = np.append(y_pred, sample_pred_new[0])
                    subject_ids_new = np.append(subject_ids, subject_ids[idx])
                    X_all = np.append(X.values, X_i, axis=0)

                    colors = [trace["marker"]["line"]["color"]] * len(y) + ["gray"]
                    widths = [trace["marker"]["line"]["width"]] * len(y) + [5]
                    sizes = [trace["marker"]["size"]] * len(y) + [20]
                else:
                    y_new = np.append(y[:-1], y[idx])
                    y_pred_new = np.append(y_pred[:-1], sample_pred_new[0])
                    subject_ids_new = np.append(subject_ids[:-1], subject_ids[idx])
                    X_all = np.append(X.values[:-1], X_i, axis=0)
                    sample_ids_new = np.append(sample_ids[:-1], sample_ids[idx])

                    colors = trace["marker"]["line"]["color"]
                    sizes = trace["marker"]["size"]

                    widths = trace["marker"]["line"]["width"]
                    fig.data = fig.data[:-1]
                customdata = [
                    (a, b, x) for a, b, x in zip(subject_ids_new, sample_ids_new, X_all)
                ]

                with fig.batch_update():
                    trace["marker"]["line"]["color"] = colors
                    trace["marker"]["size"] = sizes

                    trace["marker"]["line"]["width"] = widths
                    trace["marker"]["line"]["color"] = "black"
                    trace["customdata"] = customdata

                    trace["x"] = y_new
                    trace["y"] = y_pred_new

                    y_pred_delta = y_pred_new[-1] - y_pred[idx]  # * fig.layout.height
                    print(y_pred_delta)

                    fig.add_trace(
                        go.Scatter(
                            x=[y_new[-1], y_new[-1]],
                            y=[y_pred[idx], y_pred_new[-1]],
                            line_color="rgba(0,0,0,1)",
                            name="Intervention",
                            hoveron="points",
                            mode="lines+markers",
                            marker=dict(size=10, line=dict(width=2)),
                            marker_symbol="circle-open",
                            line={"width": 3},
                        )
                    )

                config = {
                    "toImageButtonOptions": {
                        "format": "svg",  # one of png, svg, jpeg, webp
                        "filename": "embedding_to_latent_space",
                        "height": fig["layout"]["height"],
                        "width": fig["layout"]["width"],
                        "scale": 1,  # Multiply title/legend/axis/canvas sizes by this factor
                    }
                }

            results = {
                "fig": fig,
                "trace": trace,
                "ret_val": ret_val,
                "config": config,
            }
            return results

        return _inner

    def add_longitudinal(self, fig, indices):
        """Add longitudinal analysis per SubjectID to figure.

        Parameters
        ----------
        fig : plotly.graph_objects.Figure
            Figure to which the longitudinal analysis will be added.
        indices : list
            Indices of samples from which the longitudinal analysis will be
            calculated for every subjectID.

        Returns
        -------
        fig : plotly.graph_objects.Figure
            Figure with added longitudinal analysis.
        """
        default_linewidth = 3
        highlighted_linewidth_delta = 5
        number_of_figures_start = len(fig.data)

        subject_ids = self.subject_ids[indices]
        y = self.y[indices]
        y_pred = self.y_pred[indices]

        first_iteration = True
        for subject_id in natsorted(np.unique(subject_ids)):
            indices = subject_ids == subject_id
            y_subject = y[indices]
            y_pred_subject = y_pred[indices]

            scatter_trace = dict(
                x=y_subject,
                y=y_pred_subject,
                line_color="rgba(0,0,0,0.5)",
                name=subject_id,
                hoveron="points",
                mode="lines+markers",
                marker=dict(size=10, line=dict(width=2)),
                marker_symbol="circle-open",
                line={"width": default_linewidth, "dash": "dashdot"},
                visible="legendonly",
            )
            if first_iteration:
                scatter_trace = {
                    "legendgroup": "<b>Longitudinal subjects</b>",
                    "legendgrouptitle_text": "<b>Longitudinal subjects</b>",
                    **scatter_trace,
                }
                first_iteration = False
            fig.add_trace(scatter_trace)

        number_of_figures_end = len(fig.data) - number_of_figures_start
        # our custom event handler
        def _update_trace(trace, points, selector):
            # this list stores the points which were clicked on
            # in all but one trace they are empty
            if len(points.point_inds) == 0:
                return

            for i in range(number_of_figures_start, number_of_figures_end):
                fig.data[i]["line"][
                    "width"
                ] = default_linewidth + highlighted_linewidth_delta * (
                    i == points.trace_index
                )

        # we need to add the on_click event to each trace separately
        for i in range(number_of_figures_start, number_of_figures_end):
            fig.data[i].on_click(_update_trace)

        return fig

    def add_trajectory(self, fig, indices, name, color, degree, **kwargs):
        """Plot trajectory mean and intervals.

        Parameters
        ----------
        fig : plotly.graph_objects.Figure
            Figure to which the trajectory will be added.
        indices : list
            Indices of samples from which the trajectory will be calculated.
        name : str
            Name of the trajectory.
        color : str
            Color of the trajectory.
        degree : int
            Degree of the polynomial used for the trajectory.
        **kwargs : dict
            Additional arguments for the plotly.graph_objects.Scatter object.

        Returns
        -------
        fig : plotly.graph_objects.Figure
            Figure with added trajectory.
        """
        results = self.get_prediction_and_confidence_interval(
            indices=indices,
            degree=degree,
        )

        pi_x = results["prediction_interval"]["x"]
        pi_mean = results["prediction_interval"]["mean"]
        pi_bound = results["prediction_interval"]["bound"]

        ci_x = results["confidence_interval"]["x"]
        ci_mean = results["confidence_interval"]["mean"]
        ci_bound = results["confidence_interval"]["bound"]

        # prediction interval
        fig.add_trace(
            go.Scatter(
                x=list(pi_x) + list(pi_x[::-1]),
                y=list(pi_mean - pi_bound) + list(pi_mean + pi_bound)[::-1],
                fill="toself",
                fillcolor=f"rgba({color},0.3)",
                line_color=f"rgba({color},0.5)",
                showlegend=True,
                name="95% Prediction Interval",
                hoveron="points",
                legendgroup=f"<b>{name}</b>",
                legendgrouptitle_text=f"<b>{name}</b>",
                **kwargs,
            )
        )
        # confidence interval
        fig.add_trace(
            go.Scatter(
                x=list(ci_x) + list(ci_x[::-1]),
                y=list(ci_mean - ci_bound) + list(ci_mean + ci_bound)[::-1],
                fill="toself",
                fillcolor=f"rgba({color},0.6)",
                line_color=f"rgba({color},0.8)",
                showlegend=True,
                name="95% Confidence Interval",
                hoveron="points",
                **kwargs,
            )
        )
        # mean prediction
        fig.add_trace(
            go.Scatter(
                x=pi_x,
                y=pi_mean,
                line_color=f"rgba({color},1.0)",
                name="Trajectory mean",
                hoveron="points",
                **kwargs,
            )
        )

        return fig

    def add_samples(self, fig, indices, color, **kwargs):
        """Add samples to figure.

        Parameters
        ----------
        fig : plotly.graph_objects.Figure
            Figure to which the samples will be added.
        indices : list
            Indices of samples from which the samples will be calculated.
        color : str
            Color of the samples.

        Returns
        -------
        fig : plotly.graph_objects.Figure
            Figure with added samples.
        """
        y = self.y[indices]
        y_pred = self.y_pred[indices]
        sample_ids = self.sample_ids[indices]
        subject_ids = self.subject_ids[indices]
        X = self.X.values[indices]

        # samples
        fig.add_trace(
            go.Scatter(
                x=y,
                y=y_pred,
                line_color=f"rgba({color},0.8)",
                name="Samples",
                mode="markers",
                marker=dict(size=8, line=dict(width=2, color="rgba(0.0,0.0,0.0,0.8)")),
                hoveron="points",
                customdata=[(a, b, x) for a, b, x in zip(subject_ids, sample_ids, X)],
                hovertemplate="Age: %{x:.2f}<br>MMI: %{y:.2f}"
                + "<br>Subject: %{customdata[0]}<br>Sample: %{customdata[1]}",
                **kwargs,
            )
        )
        return fig

    def add_anomalies(self, fig, color, anomaly_type, indices):
        """Add anomalies to figure.

        Parameters
        ----------
        fig : plotly.graph_objects.Figure
            Figure to which the anomalies will be added.
        color : str
            Color of the anomalies.
        anomaly_type : str
            Type of the anomalies.
        indices : list
            Indices of samples from which the anomalies will be calculated.

        Returns
        -------
        fig : plotly.graph_objects.Figure
            Figure with added anomalies.
        """
        anomalies = self.get_anomalies(anomaly_type=anomaly_type, indices=indices)

        anomaly_indices = anomalies == True
        y = self.y[anomaly_indices]
        y_pred = self.y_pred[anomaly_indices]
        sample_ids = self.sample_ids[anomaly_indices]
        subject_ids = self.subject_ids[anomaly_indices]

        fig.add_trace(
            go.Scatter(
                x=y,
                y=y_pred,
                line_color=f"rgba({color},0.7)",
                name="Anomaly",
                mode="markers",
                marker=dict(size=20, line=dict(width=5)),
                marker_symbol="cross-open",
                hoveron="points",
                customdata=[(a, b) for a, b in zip(subject_ids, sample_ids)],
                hovertemplate="Age: %{x:.2f}<br>MMI: %{y:.2f}"
                + "<br>Subject: %{customdata[0]}<br>Sample: %{customdata[1]}",
            )
        )

        return fig

    def _get_axis_max_limit(self, fig, axis_name):
        maxs = []
        for trace_data in fig.data:
            if len(trace_data[axis_name]) > 0:
                maxs.append(max(trace_data[axis_name]))
        return max(maxs)

    def plot_reference_trajectory(
        self,
        xaxis_settings=None,
        yaxis_settings=None,
        layout_settings=None,
        degree=2,
    ):
        """Plot reference samples, fit line, its CI and PI, and longitudinal data per subject.

        Parameters
        ----------
        xaxis_settings : dict
            Settings for xaxis.
        yaxis_settings : dict
            Settings for yaxis.
        layout_settings : dict
            Settings for layout.
        degree : int
            Degree of the polynomial fit.

        Returns
        -------
        results : dict
            Dictionary with the following keys:
                - "fig" : plotly.graph_objects.Figure
                    Figure with the reference samples, fit line, its CI and PI, and longitudinal data per subject.
                - "ret_val" : str
                    String with the informational text on performance.
                - "config" : dict
                    Dictionary for the plotly export image options.
        """
        indices = self.reference_groups == True

        ret_val = "<b>Performance Information</b><br>"
        ret_val += (
            f"MAE: {mean_absolute_error(self.y[indices], self.y_pred[indices]):.3f}<br>"
        )
        ret_val += f"R^2: {r2_score(self.y[indices], self.y_pred[indices]):.3f}<br>"

        if layout_settings is None:
            layout_settings = {}
        layout_settings_final = {**self.layout_settings_default, **layout_settings}
        if xaxis_settings is None:
            xaxis_settings = {}
        if yaxis_settings is None:
            yaxis_settings = {}
        xaxis_settings_final = {**self.axis_settings_default, **xaxis_settings}
        yaxis_settings_final = {**self.axis_settings_default, **yaxis_settings}

        fig = go.FigureWidget()

        fig = self.add_trajectory(
            fig=fig,
            indices=indices,
            name="Reference",
            color=self.color_reference,
            degree=degree,
        )
        fig = self.add_samples(
            fig=fig,
            indices=indices,
            color=self.color_reference,
        )

        fig.update_xaxes(
            title=f"Age at collection [{self.dataset.time_unit.name}]",
            # range=(0.0, self._get_axis_max_limit(fig, "x")),
            **xaxis_settings_final,
        )
        fig.update_yaxes(
            title=f"Microbiome Maturation Index [{self.dataset.time_unit.name}]",
            # range=(0.0, self._get_axis_max_limit(fig, "y")),
            **yaxis_settings_final,
        )
        fig.update_layout(
            title_text="Microbiome trajectory with reference samples",
            **layout_settings_final,
        )

        config = {
            "toImageButtonOptions": {
                "format": "svg",  # one of png, svg, jpeg, webp
                "filename": "microbiome_reference_trajectory",
                "height": layout_settings_final["height"],
                "width": layout_settings_final["width"],
                "scale": 1,  # Multiply title/legend/axis/canvas sizes by this factor
            }
        }

        results = {"fig": fig, "ret_val": ret_val, "config": config}

        return results

    def plot_reference_groups(
        self,
        xaxis_settings=None,
        yaxis_settings=None,
        layout_settings=None,
        degree=2,
    ):
        """Plot the reference vs non-reference trajectory.

        Parameters
        ----------
        xaxis_settings : dict
            Settings for xaxis.
        yaxis_settings : dict
            Settings for yaxis.
        layout_settings : dict
            Settings for layout.
        degree : int
            Degree of the polynomial fit.

        Returns
        -------
        results : dict
            Dictionary with the following keys:
                - "fig" : plotly.graph_objects.Figure
                    Figure with the reference samples, fit line, its CI and PI, and longitudinal data per subject.
                - "ret_val" : str
                    String with the informational text on performance.
                - "config" : dict
                    Dictionary for the plotly export image options.
        """
        assert len(self.dataset.df.reference_group.unique()) == 2, "Reference groups is available, but no samples for non-reference."
        
        ret_val = "<b>Performance Information</b><br>"
        indices = self.reference_groups == True
        _mae = mean_absolute_error(self.y[indices], self.y_pred[indices])
        _r2 = r2_score(self.y[indices], self.y_pred[indices])
        if _mae != 0 and not math.isnan(_r2):
            ret_val += f"MAE reference: {_mae:.3f}<br>"
            ret_val += f"R^2 reference: {_r2:.3f}<br>"

        indices = self.reference_groups != True
        _mae = mean_absolute_error(self.y[indices], self.y_pred[indices])
        _r2 = r2_score(self.y[indices], self.y_pred[indices])
        if _mae != 0 and not math.isnan(_r2):
            ret_val += f"MAE non-reference: {_mae:.3f}<br>"
            ret_val += f"R^2 non-reference: {_r2:.3f}<br>"

        indices = (self.reference_groups == True) | (self.reference_groups != True)

        if degree == 1:
            ret_val += "<br><b>Linear p-value (k, n)</b>:"
            pval_k, pval_n = self.get_pvalue_linear(column="reference_groups", indices=indices)
            ret_val += f"<br>Reference vs. Non-reference: ({pval_k:.3f}, {pval_n:.3f})"
        else:
            ret_val += "<b>Spline p-value</b>:"
            pval = self.get_pvalue_spline(column="reference_groups", indices=indices, degree=degree)
            ret_val += f"<br>Reference vs. Non-reference: {pval:.3f}"

        if layout_settings is None:
            layout_settings = {}
        layout_settings_final = {**self.layout_settings_default, **layout_settings}
        if xaxis_settings is None:
            xaxis_settings = {}
        if yaxis_settings is None:
            yaxis_settings = {}
        xaxis_settings_final = {**self.axis_settings_default, **xaxis_settings}
        yaxis_settings_final = {**self.axis_settings_default, **yaxis_settings}

        fig = go.FigureWidget()

        ret_val += "<br><br><b>Number of samples per reference group</b>"
        # plot reference trajectory
        indices = self.reference_groups == True
        try:
            fig = self.add_trajectory(
                fig=fig,
                indices=indices,
                name="Reference",
                color=self.color_reference,
                degree=degree,
            )
            fig = self.add_samples(
                fig=fig,
                indices=indices,
                color=self.color_reference,
            )
            ret_val += f"<br>reference: {sum(indices.astype(int))} sample(s)"
        except ValueError as e:
            print(e)
            ret_val += f"<br>reference: {sum(indices.astype(int))} sample(s) - not enough to build a trajectory"

        # plot non-reference trajectory
        indices = self.reference_groups != True
        try:
            fig = self.add_trajectory(
                fig=fig,
                indices=indices,
                name="Non-reference",
                color=self.color_non_reference,
                degree=degree,
            )
            fig = self.add_samples(
                fig=fig,
                indices=indices,
                color=self.color_non_reference,
            )
            ret_val += f"<br>non-reference: {sum(indices.astype(int))} sample(s)"
        except ValueError as e:
            print(e)
            ret_val += f"<br>non-reference: {sum(indices.astype(int))} sample(s) - not enough to build a trajectory"

        # plot longitudinal
        fig = self.add_longitudinal(fig, indices=indices)

        fig.update_xaxes(
            title=f"Age at collection [{self.dataset.time_unit.name}]",
            # range=(0.0, self._get_axis_max_limit(fig, "x")),
            **xaxis_settings_final,
        )
        fig.update_yaxes(
            title=f"Microbiome Maturation Index [{self.dataset.time_unit.name}]",
            # range=(0.0, self._get_axis_max_limit(fig, "y")),
            **yaxis_settings_final,
        )

        fig.update_layout(
            title_text="Microbiome trajectory - trajectory per reference group",
            **layout_settings_final,
        )

        config = {
            "toImageButtonOptions": {
                "format": "svg",  # one of png, svg, jpeg, webp
                "filename": "microbiome_trajectory_per_reference_group",
                "height": layout_settings_final["height"],
                "width": layout_settings_final["width"],
                "scale": 1,  # Multiply title/legend/axis/canvas sizes by this factor
            }
        }

        results = {"fig": fig, "ret_val": ret_val, "config": config}

        return results

    def plot_groups(
        self,
        xaxis_settings=None,
        yaxis_settings=None,
        layout_settings=None,
        degree=2,
    ):
        """Plot the trajectory per group.

        Parameters
        ----------
        xaxis_settings : dict
            Settings for xaxis.
        yaxis_settings : dict
            Settings for yaxis.
        layout_settings : dict
            Settings for layout.
        degree : int
            Degree of the polynomial fit.

        Returns
        -------
        results : dict
            Dictionary with the following keys:
                - "fig" : plotly.graph_objects.Figure
                    Figure with the reference samples, fit line, its CI and PI, and longitudinal data per subject.
                - "ret_val" : str
                    String with the informational text on performance.
                - "config" : dict
                    Dictionary for the plotly export image options.
        """
        assert len(self.dataset.df.group.unique()) >= 2, "There are no at least 2 groups to compare."
        group_vals = np.unique(self.groups)
        ret_val = "<b>Performance Information</b><br>"
        for group in group_vals:
            indices = self.groups == group
            _mae = mean_absolute_error(self.y[indices], self.y_pred[indices])
            _r2 = r2_score(self.y[indices], self.y_pred[indices])
            if _mae != 0 and not math.isnan(_r2):
                ret_val += f"MAE {group}: {_mae:.3f}<br>"
                ret_val += f"R^2 {group}: {_r2:.3f}<br>"
        
        comb = combinations(group_vals, 2)
        for c in list(comb):
            indices = (self.groups == c[0]) | (self.groups == c[1])
            if degree == 1:
                
                pval_k, pval_n = self.get_pvalue_linear(
                    column="groups",
                    indices=indices,
                )
                if not math.isnan(pval_k) and not math.isnan(pval_n):
                    ret_val += "<br><b>Linear p-value (k, n)</b>:"
                    ret_val += f"<br>{c[0]} vs. {c[1]}: ({pval_k:.3f}, {pval_n:.3f})"
            else:
                
                pval = self.get_pvalue_spline(
                    column="groups",
                    indices=indices,
                    degree=degree,
                )
                if not math.isnan(pval):
                    ret_val += "<br><b>Spline p-value</b>:"
                    ret_val += f"<br>{c[0]} vs. {c[1]}: {pval:.3f}"

        if layout_settings is None:
            layout_settings = {}
        layout_settings_final = {**self.layout_settings_default, **layout_settings}
        if xaxis_settings is None:
            xaxis_settings = {}
        if yaxis_settings is None:
            yaxis_settings = {}
        xaxis_settings_final = {**self.axis_settings_default, **xaxis_settings}
        yaxis_settings_final = {**self.axis_settings_default, **yaxis_settings}

        fig = go.FigureWidget()
        i = 0
        ret_val += "<br><br><b>Number of samples per group</b>"
        for group in natsorted(np.unique(self.groups)):
            indices = self.groups == group

            try:
                fig = self.add_trajectory(
                    fig=fig,
                    indices=indices,
                    name=group,
                    color=self.colors_rgb[i],
                    degree=degree,
                )
                fig = self.add_samples(
                    fig=fig,
                    indices=indices,
                    color=self.colors_rgb[i],
                )
                i += 1
                ret_val += f"<br>{group}: {sum(indices.astype(int))} sample(s)"
            except ValueError as e:
                print(e)
                ret_val += f"<br>{group}: {sum(indices.astype(int))} sample(s) - not enough to build a trajectory"

        indices = range(len(self.dataset.df))
        fig = self.add_longitudinal(
            fig=fig,
            indices=indices,
        )

        fig.update_xaxes(
            title=f"Age at collection [{self.dataset.time_unit.name}]",
            # range=(0.0, self._get_axis_max_limit(fig, "x")),
            **xaxis_settings_final,
        )
        fig.update_yaxes(
            title=f"Microbiome Maturation Index [{self.dataset.time_unit.name}]",
            # range=(0.0, self._get_axis_max_limit(fig, "y")),
            **yaxis_settings_final,
        )

        fig.update_layout(
            title_text="Microbiome trajectory - trajectory per group",
            **layout_settings_final,
        )

        config = {
            "toImageButtonOptions": {
                "format": "svg",  # one of png, svg, jpeg, webp
                "filename": "microbiome_trajectory_per_group",
                "height": layout_settings_final["height"],
                "width": layout_settings_final["width"],
                "scale": 1,  # Multiply title/legend/axis/canvas sizes by this factor
            }
        }

        results = {"fig": fig, "ret_val": ret_val, "config": config}

        return results

    def plot_anomalies(
        self,
        anomaly_type=AnomalyType.PREDICTION_INTERVAL,
        xaxis_settings=None,
        yaxis_settings=None,
        layout_settings=None,
        degree=2,
    ):
        """Plot anomalies.

        Parameters
        ----------
        anomaly_type : AnomalyType
            Type of anomaly to plot.
        xaxis_settings : dict
            Settings for xaxis.
        yaxis_settings : dict
            Settings for yaxis.
        layout_settings : dict
            Settings for layout.
        degree : int
            Degree of the polynomial fit.

        Returns
        -------
        results : dict
            Dictionary with the following keys:
                - "fig" : plotly.graph_objects.Figure
                    Figure with the reference samples, fit line, its CI and PI, and longitudinal data per subject.
                - "ret_val" : str
                    String with the informational text on performance.
                - "config" : dict
                    Dictionary for the plotly export image options.
        """
        indices = self.reference_groups == True

        ret_val = "<b>Performance Information</b><br>"
        ret_val += (
            f"MAE: {mean_absolute_error(self.y[indices], self.y_pred[indices]):.3f}<br>"
        )
        ret_val += f"R^2: {r2_score(self.y[indices], self.y_pred[indices]):.3f}<br>"

        if layout_settings is None:
            layout_settings = {}
        layout_settings_final = {**self.layout_settings_default, **layout_settings}
        if xaxis_settings is None:
            xaxis_settings = {}
        if yaxis_settings is None:
            yaxis_settings = {}
        xaxis_settings_final = {**self.axis_settings_default, **xaxis_settings}
        yaxis_settings_final = {**self.axis_settings_default, **yaxis_settings}

        fig = go.FigureWidget()

        fig = self.add_trajectory(
            fig=fig,
            indices=indices,
            name="Reference",
            color=self.color_reference,
            degree=degree,
        )
        fig = self.add_samples(
            fig=fig,
            indices=indices,
            color=self.color_reference,
        )

        fig = self.add_anomalies(
            fig=fig,
            color=self.color_anomaly,
            anomaly_type=anomaly_type,
            indices=indices,
        )

        fig.update_xaxes(
            title=f"Age at collection [{self.dataset.time_unit.name}]",
            # range=(0.0, self._get_axis_max_limit(fig, "x")),
            **xaxis_settings_final,
        )
        fig.update_yaxes(
            title=f"Microbiome Maturation Index [{self.dataset.time_unit.name}]",
            # range=(0.0, self._get_axis_max_limit(fig, "y")),
            **yaxis_settings_final,
        )
        fig.update_layout(
            title_text="Microbiome trajectory - trajectory per reference group",
            **layout_settings_final,
        )

        config = {
            "toImageButtonOptions": {
                "format": "svg",  # one of png, svg, jpeg, webp
                "filename": "microbiome_reference_trajectory",
                "height": layout_settings_final["height"],
                "width": layout_settings_final["width"],
                "scale": 1,  # Multiply title/legend/axis/canvas sizes by this factor
            }
        }

        results = {"fig": fig, "ret_val": ret_val, "config": config}

        return results

    def plot_timeboxes(
        self,
        time_block_ranges,
        anomaly_type=None,
        xaxis_settings=None,
        yaxis_settings=None,
        layout_settings=None,
        num_top_bacteria=5,
        degree=2,
    ):
        """Plot importance time boxes for reference non-anomalous samples.

        Parameters
        ----------
        time_block_ranges : list
            List of time units to plot.
        anomaly_type : AnomalyType
            Type of anomaly to plot.
        xaxis_settings : dict
            Settings for xaxis.
        yaxis_settings : dict
            Settings for yaxis.
        layout_settings : dict
            Settings for layout.
        degree : int
            Degree of the polynomial fit.
        num_top_bacteria : int
            Number of top bacteria to plot per block.

        Returns
        -------
        results : dict
            Dictionary with the following keys:
                - "fig" : plotly.graph_objects.Figure
                    Figure with the reference samples, fit line, its CI and PI, and longitudinal data per subject.
                - "ret_val" : str
                    String with the informational text on performance.
                - "config" : dict
                    Dictionary for the plotly export image options.
        """
        indices = self.reference_groups == True  # & (self.anomaly == False)
        anomaly_type = anomaly_type or self.anomaly_type

        ret_val = "<b>Performance Information</b><br>"
        ret_val += (
            f"MAE: {mean_absolute_error(self.y[indices], self.y_pred[indices]):.3f}<br>"
        )
        ret_val += f"R^2: {r2_score(self.y[indices], self.y_pred[indices]):.3f}<br>"

        if layout_settings is None:
            layout_settings = {}
        layout_settings_final = {**self.layout_settings_default, **layout_settings}
        if xaxis_settings is None:
            xaxis_settings = {}
        if yaxis_settings is None:
            yaxis_settings = {}
        xaxis_settings_final = {**self.axis_settings_default, **xaxis_settings}
        yaxis_settings_final = {**self.axis_settings_default, **yaxis_settings}

        fig = go.FigureWidget()

        fig = self.add_trajectory(
            fig=fig,
            indices=indices,
            name="Reference",
            color=self.color_reference,
            degree=degree,
            hoverinfo="skip",
        )
        fig = self.add_samples(
            fig=fig,
            indices=indices,
            color=self.color_reference,
            hoverinfo="skip",
        )

        anomalies = self.get_anomalies(anomaly_type=anomaly_type, indices=indices)
        indices = np.logical_and(indices, anomalies == False)

        fig = self.add_timeboxes(
            fig,
            time_block_ranges,
            indices=indices,
            num_top_bacteria=num_top_bacteria,
        )

        fig.update_xaxes(
            title=f"Age at collection [{self.dataset.time_unit.name}]",
            **xaxis_settings_final,
        )
        fig.update_yaxes(
            title=f"Microbiome Maturation Index [{self.dataset.time_unit.name}]",
            **yaxis_settings_final,
        )
        fig.update_layout(
            title_text="Microbiome trajectory - importance timeboxes",
            **layout_settings_final,
        )

        config = {
            "toImageButtonOptions": {
                "format": "svg",  # one of png, svg, jpeg, webp
                "filename": "microbiome_importance_timeboxes",
                "height": layout_settings_final["height"],
                "width": layout_settings_final["width"],
                "scale": 1,  # Multiply title/legend/axis/canvas sizes by this factor
            }
        }

        results = {"fig": fig, "ret_val": ret_val, "config": config}

        return results

    def plot_intervention(
        self,
        time_block_ranges,
        anomaly_type=None,
        xaxis_settings=None,
        yaxis_settings=None,
        layout_settings=None,
        num_top_bacteria=5,
        degree=2,
    ):
        """Plot interactive sample intervention that returns samples back to the non-anomalous referene samples.

        Parameters
        ----------
        time_block_ranges : list
            List of time units to plot.
        anomaly_type : AnomalyType
            Type of anomaly to plot.
        xaxis_settings : dict
            Settings for xaxis.
        yaxis_settings : dict
            Settings for yaxis.
        layout_settings : dict
            Settings for layout.
        num_top_bacteria : int
            Number of top bacteria to plot per block.
        degree : int
            Degree of the polynomial fit.

        Returns
        -------
        results : dict
            Dictionary with the following keys:
                - "fig" : plotly.graph_objects.Figure
                    Figure with the reference samples, fit line, its CI and PI, and longitudinal data per subject.
                - "ret_val" : str
                    String with the informational text on performance.
                - "config" : dict
                    Dictionary for the plotly export image options.
        """
        indices = self.reference_groups == True
        anomaly_type = anomaly_type or self.anomaly_type

        ret_val = "<b>Performance Information</b><br>"
        ret_val += (
            f"MAE: {mean_absolute_error(self.y[indices], self.y_pred[indices]):.3f}<br>"
        )
        ret_val += f"R^2: {r2_score(self.y[indices], self.y_pred[indices]):.3f}<br>"

        if layout_settings is None:
            layout_settings = {}
        layout_settings_final = {**self.layout_settings_default, **layout_settings}
        if xaxis_settings is None:
            xaxis_settings = {}
        if yaxis_settings is None:
            yaxis_settings = {}
        xaxis_settings_final = {**self.axis_settings_default, **xaxis_settings}
        yaxis_settings_final = {**self.axis_settings_default, **yaxis_settings}

        fig = go.FigureWidget()

        fig = self.add_trajectory(
            fig=fig,
            indices=indices,
            name="Reference",
            color=self.color_reference,
            degree=degree,
        )
        fig = self.add_samples(
            fig=fig,
            indices=indices,
            color=self.color_reference,
        )

        anomalies = self.get_anomalies(anomaly_type=anomaly_type, indices=indices)
        intervention_indice = np.logical_and(indices, anomalies == False)

        fig = self.add_timeboxes(
            fig,
            time_block_ranges,
            indices=intervention_indice,
            num_top_bacteria=num_top_bacteria,
        )

        fig = self.add_intervention(
            fig=fig,
            time_block_ranges=time_block_ranges,
            indices=intervention_indice,
        )

        fig.update_xaxes(
            title=f"Age at collection [{self.dataset.time_unit.name}]",
            **xaxis_settings_final,
        )
        fig.update_yaxes(
            title=f"Microbiome Maturation Index [{self.dataset.time_unit.name}]",
            **yaxis_settings_final,
        )
        fig.update_layout(
            title_text="Microbiome trajectory - trajectory per reference group",
            **layout_settings_final,
        )

        config = {
            "toImageButtonOptions": {
                "format": "svg",  # one of png, svg, jpeg, webp
                "filename": "microbiome_reference_trajectory",
                "height": layout_settings_final["height"],
                "width": layout_settings_final["width"],
                "scale": 1,  # Multiply title/legend/axis/canvas sizes by this factor
            }
        }

        results = {"fig": fig, "ret_val": ret_val, "config": config}

        return results

    def plot_animated_longitudinal_information(
        self,
        xaxis_settings=None,
        yaxis_settings=None,
        layout_settings=None,
        degree=2,
    ):
        """Plot animated longitudinal information of reference samples.

        Parameters
        ----------
        xaxis_settings : dict
            Settings for xaxis.
        yaxis_settings : dict
            Settings for yaxis.
        layout_settings : dict
            Settings for layout.

        Returns
        -------
        results : dict
            Dictionary with the following keys:
                - "fig" : plotly.graph_objects.Figure
                    Figure with the interactive longitudinal information.
        """
        indices = self.reference_groups == True

        if layout_settings is None:
            layout_settings = {}
        layout_settings_final = {**self.layout_settings_default, **layout_settings}
        if xaxis_settings is None:
            xaxis_settings = {}
        if yaxis_settings is None:
            yaxis_settings = {}
        xaxis_settings_final = {**self.axis_settings_default, **xaxis_settings}
        yaxis_settings_final = {**self.axis_settings_default, **yaxis_settings}

        frames = []

        trajectory = self.add_trajectory(
            fig=go.FigureWidget(),
            indices=indices,
            name="Reference",
            color="220,220,220",
            degree=degree,
        ).data

        subject_ids = self.subject_ids[indices]
        sample_ids = self.sample_ids[indices]
        y = self.y[indices]
        y_pred = self.y_pred[indices]

        first_iteration = True
        for age_at_collection in np.unique(y):
            data = [*trajectory]
            indices_time = y <= age_at_collection

            for subject_id in natsorted(np.unique(subject_ids)):
                indices_subject = list(subject_ids[indices_time] == subject_id)

                y_subject = y[indices_time][indices_subject]
                y_pred_subject = y_pred[indices_time][indices_subject]
                sample_ids_subject = sample_ids[indices_time][indices_subject]

                scatter_trace = dict(
                    x=y_subject,
                    y=y_pred_subject,
                    line_color=f"rgba({self.get_subject_colors(subject_id)},0.9)",
                    name=subject_id,
                    mode="lines+markers",
                    marker=dict(size=10, line=dict(width=3)),
                    marker_symbol="circle-open",
                    hoveron="points",
                    customdata=sample_ids_subject,
                    hovertemplate="Age: %{x:.2f}<br>MMI: %{y:.2f}"
                    + f"<br>Subject: {subject_id}<br>"
                    + "Sample: %{customdata}",
                )
                if first_iteration:
                    scatter_trace = {
                        "legendgroup": "<b>Longitudinal subjects</b>",
                        "legendgrouptitle_text": "<b>Longitudinal subjects</b>",
                        **scatter_trace,
                    }
                    first_iteration = False
                data.append(scatter_trace)
            frames.append(go.Frame(data=data))

        fig = go.Figure(
            data=frames[0]["data"],
            layout=go.Layout(
                updatemenus=[
                    dict(
                        type="buttons",
                        buttons=[dict(label="Play", method="animate", args=[None])],
                    )
                ]
            ),
            frames=frames,
        )

        fig.update_xaxes(
            title=f"Age at collection [{self.dataset.time_unit.name}]",
            **xaxis_settings_final,
        )
        fig.update_yaxes(
            title=f"Microbiome Maturation Index [{self.dataset.time_unit.name}]",
            **yaxis_settings_final,
        )
        fig.update_layout(
            title_text="Animated longitudinal information",
            **layout_settings_final,
        )

        config = {
            "toImageButtonOptions": {
                "format": "svg",  # one of png, svg, jpeg, webp
                "filename": "animation_longitudinal_information",
                "height": layout_settings_final["height"],
                "width": layout_settings_final["width"],
                "scale": 1,  # Multiply title/legend/axis/canvas sizes by this factor
            }
        }

        results = {"fig": fig, "ret_val": None, "config": config}

        return results



"""
Implementation of different statistical analysis tools that are used to compare the significant difference between two trajectories.

(1) Splinectomy longitudinal statistical analysis tools 
References: 
- https://github.com/RRShieldsCutler/splinectomeR
- https://www.frontiersin.org/articles/10.3389/fmicb.2018.00785/full

The following 3 methods below are translated from R to Python and acomodated for our project.
Original package is called splinectomeR, implemented in R. For more details please use the reference links.
In short, the test compares whether the area between two polynomial lines is significantly different.
Our trajectory lines are the plynomial lines with degree 2 or 3.

(2) Linear regression statistical analysis tools
After the 3 methods from splinectomeR, we have the other one, based on comparing the two linear lines (line y=k*x+n).
To compare two linear lines, we compare the significant difference in the two coefficients that represent the line k and n.
In the microbiome trajectory, the only linear part is the part where infant is younger than 2 years. After that, the
trajectory plateaus, therefore, this test can be used only before 2 years. 
"""
import matplotlib
matplotlib.use("agg")
import numpy as np
import pandas as pd
import seaborn as sns
import statsmodels.api as sm
from matplotlib.figure import Figure


def permuspliner(
    data,
    xvar,
    yvar,
    category,
    degree,
    cases,
    groups=None,
    perms=999,
    test_direction="more",
    ints=1000,
    quiet=False,
    cut_sparse=4,
):
    """Permuspliner tests for a significant difference between two groups overall.

    Permutation to test whether there is a non-zero trend among a set
    of individuals/samples over a continuous variable -- such as time. So, there
    does not need to be two groups in this test. The x variable datapoints are
    permuated within each case/individual, thus maintaining the distribution in
    the y component but shuffling the hypothesized trend.

    Note: the description is taken from splinectomeR

    Parameters
    ----------
    data: pd.DataFrame
        A dataframe object containing your data.
    xvar: str
        The independent variable; is continuous, e.g. y.
    yvar: str
        The dependent variable; is continuous, e.g. y_pred.
    category: str
        The column name of the category to be tested, e.g. country.
    cases:
        The column name defining the individual cases, e.g. subjectID.
    groups: dict
        If more than two groups, the two groups to compare as character vector, e.g. ["RUS", "FIN"].
    perms: int
        The number of permutations to generate
    cut_sparse: int
        Minimum number of total observations necessary per group to fit a spline (default 4)
    ints: int
        Number of x intervals over which to measure area
    quiet: bool
        Silence all text outputs

    Returns
    -------
    result: dict
        Returns a bunc of stuff, but the most important for you is pval.
    """
    perm_output = {}

    equation = lambda a, b: np.polyval(a, b)

    if test_direction not in ["more", "less"]:
        raise Exception(
            "Error in test direction option: must be either 'more' or 'less'"
        )

    in_df = data.copy()

    # Determine the two groups to compare
    if groups is None:
        if len(in_df.category.unique()) > 2:
            raise Exception(
                'More than two groups in category column. Define groups with (groups = c("Name1","Name2"))'
            )
        else:
            cats = in_df.category.unique()
            v1 = cats[0]
            v2 = cats[1]
    else:
        v1 = groups[0]
        v2 = groups[1]

    if quiet == False:
        print(f"\nGroups detected:", v1, "and", v2, ".\n")
        print(
            f"\nNow testing between variables",
            v1,
            "and",
            v2,
            "for a difference in the response labeled",
            yvar,
            "\n",
        )
        print(
            f"\nScalpel please: performing permusplinectomy with",
            perms,
            "permutations...\n",
        )

    # The experimentally reported response
    df_v1 = in_df[(in_df[category] == v1) & (~in_df[xvar].isna())]
    df_v2 = in_df[(in_df[category] == v2) & (~in_df[xvar].isna())]

    if len(df_v1[xvar]) < cut_sparse or len(df_v2[xvar]) < cut_sparse:
        raise Exception("Not enough data in each group to fit spline")

    # Prevent issues arising from identical case labels across groups
    if len(list(set(df_v1[cases]).intersection(set(df_v2[cases])))) > 0:
        raise Exception(
            "\nIt appears there may be identically labeled cases in both groups.\n...Please ensure that the cases are uniquely labeled between the two groups\n"
        )

    # Fit the splines for each group
    p_v1, _ = np.polyfit(
        df_v1[xvar], df_v1[yvar], degree, cov=True
    )  # parameters and covariance from of the fit of 1-D polynom.
    p_v2, _ = np.polyfit(
        df_v2[xvar], df_v2[yvar], degree, cov=True
    )  # parameters and covariance from of the fit of 1-D polynom.

    x0 = max(min(df_v1[xvar]), min(df_v2[xvar]))
    x1 = min(max(df_v1[xvar]), max(df_v2[xvar]))
    x0 = x0 + (
        (x1 - x0) * 0.1
    )  # Trim the first and last 10% to avoid low-density artifacts
    x1 = x1 - ((x1 - x0) * 0.1)
    xby = (x1 - x0) / (ints - 1)
    xx = np.arange(x0, x1, xby)  # Set the interval range

    # var1 , x
    v1_spl_f = pd.DataFrame(data={"var1": equation(p_v1, xx), "x": xx})
    v2_spl_f = pd.DataFrame(data={"var2": equation(p_v2, xx), "x": xx})

    real_spl_dist = (
        v1_spl_f.set_index("x").join(v2_spl_f.set_index("x"), on="x").reset_index()
    )
    real_spl_dist["abs_dist"] = np.abs(
        real_spl_dist.var1 - real_spl_dist.var2
    )  # Measure the real group distance
    real_area = (
        np.sum(real_spl_dist.abs_dist) / ints
    )  # Calculate the area between the groups

    if quiet == False:
        print(
            "\nArea between groups successfully calculated, now spinning up permutations...\n"
        )

    # Define the permutation function
    case_shuff = "case_shuff"  # Dummy label

    def spline_permute(randy, ix, perm_output):
        randy_meta = randy.drop_duplicates(subset=cases, keep="first").copy(
            deep=False
        )  # Pull out the individual IDs
        randy_meta[case_shuff] = np.random.choice(
            randy_meta[category].values,
            size=len(randy_meta[category].values),
            replace=True,
            p=None,
        )
        randy_meta = randy_meta[[cases, case_shuff]]
        randy = (
            randy.set_index(cases)
            .join(randy_meta.set_index(cases), on=cases)
            .reset_index()
        )

        randy_v1 = randy[(randy[case_shuff] == v1) & (~randy[xvar].isna())]
        randy_v2 = randy[(randy[case_shuff] == v2) & (~randy[xvar].isna())]

        # Fit the splines for the permuted groups
        p_randy_v1, _ = np.polyfit(randy_v1[xvar], randy_v1[yvar], degree, cov=True)
        p_randy_v2, _ = np.polyfit(randy_v2[xvar], randy_v2[yvar], degree, cov=True)

        # var1 , x
        randy_v1_fit = pd.DataFrame(data={"var1": equation(p_randy_v1, xx), "x": xx})
        randy_v2_fit = pd.DataFrame(data={"var2": equation(p_randy_v2, xx), "x": xx})

        spl_dist = (
            randy_v1_fit.set_index("x")
            .join(randy_v2_fit.set_index("x"), on="x")
            .reset_index()
        )
        spl_dist["abs_dist"] = np.abs(
            spl_dist.var1 - spl_dist.var2
        )  # Measure the real group distance

        real_area = (
            np.sum(real_spl_dist.abs_dist) / ints
        )  # Calculate the area between the groups

        transfer_perms = spl_dist[["var1", "var2", "abs_dist"]]
        transfer_perms = transfer_perms.rename(
            columns={
                "var1": f"v1perm_{ix}",
                "var2": f"v2perm_{ix}",
                "abs_dist": f"pdistance_{ix}",
            }
        )
        if ix > 0:
            perm_retainer = perm_output["perm_retainer"]
        else:
            perm_retainer = pd.DataFrame({"x": xx})
        perm_retainer = pd.concat([perm_retainer, transfer_perms], axis=1)
        perm_output["perm_retainer"] = perm_retainer
        perm_area = [
            np.nansum(spl_dist["abs_dist"]) / np.sum(~spl_dist["abs_dist"].isna())
        ]  # Calculate the area between permuted groups
        if ix > 0:
            permuted = perm_output["permuted"]
        else:
            permuted = []
        permuted = np.concatenate([permuted, perm_area])
        perm_output["permuted"] = permuted
        return perm_output

    # Run the permutation over desired number of iterations
    in_rand = pd.concat([df_v1, df_v2])

    pval = None

    for ix in range(perms):
        perm_output = spline_permute(in_rand, ix, perm_output)

    if quiet == False:
        print("...permutations completed...\n")

    if test_direction == "more":
        pval = (np.sum(perm_output["permuted"] >= real_area) + 1) / (perms + 1)

    elif test_direction == "less":
        pval = (sum(perm_output["permuted"] <= real_area) + 1) / (perms + 1)

    if quiet == False:
        print("p-value: ", pval)

    result = {
        "pval": pval,
        "category_1": v1,
        "category_2": v2,
        "v1_interpolated": v1_spl_f,
        "v2_interpolated": v2_spl_f,
        # "v1_spline": df_v1_spl, "v2_spline": df_v2_spl,
        "permuted_splines": perm_output["perm_retainer"],
        "true_distance": real_spl_dist,
        "v1_data": df_v1,
        "v2_data": df_v2,
    }

    return result


def permuspliner_plot_permdistance(data, xlabel=None, ylabel=""):
    """Plot permuted distance distribution

    Compare the permuted distances to the true distance. Requires permuspliner run with the "retain_perm" option.

    Parameters
    ----------
    data: pd.DataFrame
        The results object from the permuspliner function
    xlabel: str
        Name for the x axis in the plot
    ylabel: str
        Name for the y axis in the plot
    """
    sns.set_style("whitegrid")

    if xlabel is None:
        xlabel = "longitudinal parameter"

    dists = data["permuted_splines"]

    cols = dists.columns[dists.columns.str.startswith("pdistance_")]
    dists = dists.set_index("x")
    dists = dists[cols]
    num_perms = dists.shape[1]
    dists = dists.reset_index()
    dists = pd.melt(
        dists, id_vars="x", var_name="permutation", value_name="permuted_distance"
    )

    if num_perms > 1000:
        alpha_level = 0.002
    elif num_perms >= 100:
        alpha_level = 0.02
    elif num_perms < 100:
        alpha_level = 0.25

    fig = Figure(figsize=(10, 7), dpi=200)
    ax = fig.subplots()
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    for p in dists.permutation.unique():
        ax.plot(
            dists[dists.permutation == p]["x"].values,
            dists[dists.permutation == p]["permuted_distance"].values,
            alpha=alpha_level,
            solid_capstyle="butt",
            c="k",
        )

    true_dist = data["true_distance"]
    ax.plot(true_dist["x"].values, true_dist["abs_dist"].values, color="r")
    fig.show()


def permuspliner_plot_permsplines(
    data=None, xvar=None, yvar=None, xlabel="", ylabel=""
):
    """Plot permuted splines along the real data

    Compare how the permuted splines fit with the real data. Provides visual support for p values.

    Parameters
    ----------
    data: pd.DataFrame
        The results object from the permuspliner function
    xvar: str
        Name (as string) of the longitudinal x variable in the data
    yvar: str
        Name (as string) of the response/y variable in the data
    xlabel: str
        Name for the x axis in the plot
    ylabel: str
        Name for the y axis in the plot
    """
    sns.set_style("whitegrid")
    permsplines = data["permuted_splines"]
    cols = permsplines.columns[permsplines.columns.str.contains("perm")]
    permsplines = permsplines.set_index("x")
    permsplines = permsplines[cols]
    permsplines = permsplines.reset_index()

    num_perms = permsplines.shape[1] / 2
    permsplines = pd.melt(
        permsplines, id_vars="x", var_name="permutation", value_name="y.par"
    )

    var_1 = data["category_1"]
    var_2 = data["category_2"]

    true_v1 = data["v1_interpolated"]
    true_v1["Group"] = var_1
    true_v2 = data["v2_interpolated"]
    true_v2["Group"] = var_2

    true_data = pd.concat([true_v1, true_v2], axis=0)
    true_data = true_data.sort_values(by="Group")

    if num_perms > 1000:
        alpha_level = 0.002
    elif num_perms >= 100:
        alpha_level = 0.01
    elif num_perms < 100:
        alpha_level = 0.31

    fig = Figure(figsize=(10, 7), dpi=200)
    ax = fig.subplots()
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    for p in permsplines.permutation.unique():
        ax.plot(
            permsplines[permsplines.permutation == p]["x"].values,
            permsplines[permsplines.permutation == p]["y.par"].values,
            alpha=alpha_level,
            solid_capstyle="butt",
            c="k",
        )

    true_data["y"] = true_data.apply(
        lambda row: row["var1"] if row["Group"] == var_1 else row["var2"], axis=1
    )
    sns.lineplot(
        x="x", y="y", data=true_data, hue="Group", ci=None, palette=["r", "b"], ax=ax
    )
    fig.show()


##############


def regliner(df2countries_stats, mapping):
    """Linear regression statistical analysis tools

    Parameters
    ----------
    df2countries_stats: pd.DataFrame
        Dataframe containing following: Input, Output, Condition
    mapping: dict
        The two groups which effects we want to compare.

    Returns
    -------
    : float, float
        p-values of k and n respectively, representing whether the two coefficients are significantly different or not

    Examples
    --------
    >> mapping = {"RUS": 0, "FIN": 1}
    >> df2countries_stats = pd.DataFrame(data={"Input":y,
    >> ...                                     "Output":y_pred,
    >> ...                                     "Condition": df2countries.country})
    >> regliner(df2countries_stats, mapping)
    """
    dfts = df2countries_stats.copy()
    dfts["Input*Condition"] = dfts.apply(
        lambda x: mapping[x["Condition"]] * x["Input"], axis=1
    )
    dfts["Condition"] = dfts.apply(lambda x: mapping[x["Condition"]], axis=1)

    cont_indep_var = dfts["Input"]
    main_effect = dfts["Condition"]
    interaction_effect = dfts["Input*Condition"]
    outp = dfts["Output"]
    inp = np.array(dfts[["Input", "Condition", "Input*Condition"]])

    X2 = sm.add_constant(inp)
    est = sm.OLS(outp, X2)
    est2 = est.fit()

    result = {
        "k": est2.pvalues["x3"],
        "n": est2.pvalues["x2"],
        "const": est2.pvalues["const"],
    }

    return result



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="microbiome-toolbox", 
    version="1.0.2",
    author="Jelena Banjac, Shaillay Kumar Dogra, Norbert Sprenger",
    author_email="msjelenabanjac@gmail.com, ShaillayKumar.Dogra@rd.nestle.com, norbert.sprenger@rdls.nestle.com",
    maintainer="Jelena Banjac",
    maintainer_email="msjelenabanjac@gmail.com",
    description="Microbiome Toolbox",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/JelenaBanjac/microbiome-toolbox",
    packages=setuptools.find_packages(include=['microbiome', 'microbiome.*']),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=[
        "seaborn",
        "ipyvolume",
        "xgboost",
        "certifi",
        "cycler",
        "matplotlib",  #==2.2.2
        "pyparsing",
        "ipykernel",
        "ipython",
        "ipython_genutils",
        "jedi",
        "joblib",
        "jupyter_client",
        "jupyter_core",
        "numpy",
        "pandas",
        "parso",
        "pickleshare",
        "pip",
        "prompt-toolkit",
        "pygments",
        "python-dateutil",
        "scikit-learn",
        #"scikit-bio",
        "scipy",
        "setuptools",
        "six",
        "threadpoolctl",
        "tornado",
        "traitlets",
        "wheel",
        "dash==2.0.0",
        "diskcache",
        "multiprocess",
        "psutil",
        "dash-renderer",
        "dash-table",
        "dash-uploader",
        "dash-extensions",
        "plotly",
        "flask-caching",
        # "catboost",
        "statsmodels",
        "numpy",
        "shap",
        "tk",
        "gunicorn",
        "diskcache",
        "python-dotenv",
        "psutil",
        "diskcache",
        "multiprocess",
        "natsort",
        "umap-learn",
        "black",
        "isort",
        #"flake8",
    ]
)



import copy
import itertools
import math
import os
import pickle
import re
from collections import Counter
from itertools import tee
from textwrap import wrap
import matplotlib
matplotlib.use("agg")
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import plotly.express as px
import plotly.graph_objects as go
import shap
from ipywidgets import VBox
from plotly.subplots import make_subplots
from plotly.tools import mpl_to_plotly
from sklearn.decomposition import PCA
from sklearn.ensemble import RandomForestClassifier
from sklearn.metrics import confusion_matrix
from sklearn.model_selection import (
    GroupShuffleSplit,
    cross_val_predict,
    cross_val_score,
)
from sklearn.neighbors import LocalOutlierFactor

from microbiome.enumerations import (
    FeatureColumnsType,
    Normalization,
    ReferenceGroup,
    TimeUnit,
)

RANDOM_STATE = 42


class MicrobiomeDataset:

    # initial columns, necessary for the toolbox to work
    specific_columns = [
        "sampleID",
        "subjectID",
        "group",
        "age_at_collection",
        "reference_group",
    ]
    # new columns automatically created
    # Note: e.g. reference group = healthy samples
    # non reference group = non healthy samples
    # group = country
    future_columns = []

    def __init__(self, file_name=None, feature_columns=FeatureColumnsType.BACTERIA):

        if file_name == "mouse_data":
            file_name = "https://raw.githubusercontent.com/JelenaBanjac/microbiome-toolbox/main/notebooks/Mouse_16S/INPUT_FILES/website_mousedata.csv"
        elif file_name == "human_data":
            file_name = "https://raw.githubusercontent.com/JelenaBanjac/microbiome-toolbox/main/notebooks/Human_Subramanian/INPUT_FILES/subramanian_et_al_l2_ELM_website.csv"
        elif file_name is None:
            raise Exception("Specify a valid file name. Or choose existing options: `mouse_data` or `human_data`")

        # create dataframe regardless of delimiter (sep)
        self.df = pd.read_csv(file_name, sep=None, engine="python")

        print("columns", self.df.columns)

        # collect all columns that are missing
        missing_columns = []

        for col in self.specific_columns:
            if col not in self.df.columns:
                if col == "reference_group":
                    # if this column doesn't exist, create it s.t.
                    # all samples are in reference by default
                    self.df["reference_group"] = True
                elif col == "group":
                    # if this column doesn't exist, create it s.t.
                    # all samples are in one group by default
                    self.df["group"] = 'none'
                else:
                    missing_columns.append(col)

        # check if there is at least one column with bacteria
        bacteria_columns = self.df.columns[
            # (~self.df.columns.isin(self.specific_columns))
            (~self.df.columns.str.contains('|'.join(self.specific_columns)))
            & (~self.df.columns.isin(self.future_columns))
            & (~self.df.columns.str.startswith("meta_"))
            & (~self.df.columns.str.startswith("id_"))
        ].tolist()
        if len(bacteria_columns) > 0:
            # if there are bacteria columns, make sure these columns have `bacteria_` prefix
            self.df = self.df.rename(
                mapper={
                    k: f"bacteria_{k}"
                    for k in bacteria_columns
                    if not k.startswith("bacteria_")
                },
                axis=1,
            )
        else:
            missing_columns.append("bacteria_*")

        print("bacteria_columns", self.bacteria_columns)
        if len(missing_columns) > 0:
            raise Exception(
                "There was an error processing this file! Missing columns: "
                + ",".join(missing_columns)
            )
        
        self.df = self.df.apply(lambda row: self._fix_zeros(row), axis=1)

        # preprocess df
        self.df.sampleID = self.df.sampleID.astype(str)
        self.df.subjectID = self.df.subjectID.astype(str)
        self.df.reference_group = self.df.reference_group.fillna(False).astype(bool)

        self.df = self.df.convert_dtypes()

        # convert metadata string columns to categorical
        for column in self.metadata_columns:
            self.df[column] = self.df[column].astype("category")

        # dummy encoding of metadata columns (string and object type)
        self.df = pd.get_dummies(self.df, dummy_na=True)
        self.df = self.df.fillna(0)
        self.df = self.df.sort_values(by="age_at_collection", ignore_index=True)

        # initialize trajectory values
        self.df["MMI"] = 0

        # initialize feature columns (only once, in constructor!)
        if isinstance(feature_columns, (list, np.ndarray)):
            self._feature_columns = feature_columns
        elif feature_columns == FeatureColumnsType.BACTERIA:
            self._feature_columns = self.bacteria_columns
        elif feature_columns == FeatureColumnsType.METADATA:
            self._feature_columns = self.metadata_columns
        elif feature_columns == FeatureColumnsType.BACTERIA_AND_METADATA:
            self._feature_columns = self.bacteria_and_metadata_columns

        # save initial
        self.__age_at_collection = self.df.age_at_collection.to_numpy()
        self.__reference_group = self.df.reference_group.to_numpy()
        self.__df = self.df.copy(deep=True)

        # normalization by log-ratio
        self.log_ratio_bacteria = None
        # normalization by mean and std (along columns)
        self.normalized = Normalization.NON_NORMALIZED

        self.time_unit = TimeUnit.DAY
        self.set_reference_group_choice(ReferenceGroup.USER_DEFINED)
        self.nice_name = (
            lambda x: re.sub(" +", "|", re.sub("[kpcofgs]__|\.|_", " ", x[9:]).strip())
            if x.startswith("bacteria_")
            else x
        )

        self.layout_settings_default = dict(
            height=900,
            width=1200,
            plot_bgcolor="rgba(255,255,255,255)",
            paper_bgcolor="rgba(255,255,255,255)",
            margin=dict(l=70, r=70, t=70, b=70),
            font=dict(size=17),
            hoverdistance=-1,
            legend=dict(
                x=1.01,
                y=1,
            ),
        )

        self.axis_settings_default = dict(
            tick0=0,
            mirror=True,
            showline=True,
            linecolor="lightgrey",
            gridcolor="lightgrey",
            zeroline=True,
            zerolinecolor="lightgrey",
            showspikes=True,
            spikecolor="gray",
            autorange=True,
        )

    def __str__(self):
        ret_val = ""

        # dataframe size
        ret_val += "### Dataset size\n"
        ret_val += f"#samples = {self.df.shape[0]}, #columns = {self.df.shape[1]}\n"

        # counts per column type
        ret_val += "### Counts per column type\n"
        ret_val += f"Number of bacteria columns: {len(self.bacteria_columns)}\n"
        ret_val += f"Number of metadata columns: {len(self.metadata_columns)}\n"
        ret_val += f"Number of ID columns: {len(self.id_columns)}\n"
        ret_val += f"Number of other columns: {len(self.other_columns)}\n"

        # count reference
        ret_val += "### Reference group\n"
        ref_counts = self.df.reference_group.value_counts()
        ret_val += f"Reference group count vs. non-reference: {ref_counts[True]} vs. {ref_counts[False]}\n"

        return ret_val

    def _fix_zeros(self, row):
        for col in self.bacteria_columns:
            try:
                row[col] = 1e-10 if row[col] == 0.0 or row[col] < 1e-10 else row[col]
            except TypeError as e:
                raise Exception(
                    f"Check yout bacteria columns, {col[9:]} doesn't have a numerical values."
                ) from e
        return row

    @property
    def time_unit(self):
        return self._time_unit

    @time_unit.setter
    def time_unit(self, val):
        """Modify age_at_collection column"""
        if val not in TimeUnit:
            raise ValueError(f"There is no value {val} in TimeUnit enumeration class!")
        self._time_unit = val
        self.df.age_at_collection = self.__age_at_collection / val.value
        self.df.MMI = self.df.MMI / val.value

    @property
    def nice_name(self):
        return self._nice_name

    @nice_name.setter
    def nice_name(self, val):
        self._nice_name = val

    # @property
    def get_reference_group_choice(self):
        return self._reference_group_choice

    # @reference_group_choice.setter
    def set_reference_group_choice(self, val, novelty_settings=None):
        """Modify reference_group column"""
        if val not in ReferenceGroup:
            raise ValueError(
                f"There is no value {val} in ReferenceGroup enumeration class!"
            )
        self._reference_group_choice = val

        self.df.reference_group = self.__reference_group
        # TODO: feature_columns can be 3 cases
        X = self.df[self.feature_columns].values
        y = self.df.reference_group.values
        groups = self.df.subjectID.values

        if val == ReferenceGroup.NOVELTY_DETECTION:
            reference_group = self._find_best_novelty_reference_group(X, y, groups, settings=novelty_settings)
        elif val == ReferenceGroup.USER_DEFINED:
            reference_group = self.__reference_group
        else:
            raise NotImplementedError(f"Not implemented yet {val}, {type(val)}!")

        self.df.reference_group = reference_group

        # get latest values for y and groups
        y = self.df.reference_group.values
        # groups = self.df.subjectID.values
        if len(self.df.reference_group.unique()) == 1:
            results = {
                "accuracy": "",
                "f1score": "",
                "fig": None,
                "config": None,
                "img_src": None,
            }
        else:
            results = two_groups_differentiation(
                self.df[self.feature_columns],
                self.df.reference_group,
                groups,
                y_column="reference_group",
                plot=True,
            )
        self._differentiation_score = results["f1score"]

        self.reference_group_plot = results["fig"]
        self.reference_group_img_src = results["img_src"]
        self.reference_group_accuracy = results["accuracy"]
        self.reference_group_f1score = results["f1score"]
        self.reference_group_config = results["config"]

    def _find_best_novelty_reference_group(self, X, y, groups, settings=None):

        # if novelty_settings is None:
        #     c = Counter(groups)
        #     n_neighbors_min = 1
        #     n_neighbors_max = c.most_common(1)[0][1]
        #     n_neighbors_list = set(
        #         [
        #             n_neighbors_min,
        #             (n_neighbors_min + n_neighbors_max) // 2 + 1,
        #             n_neighbors_max,
        #         ]
        #     )

        #     novelty_settings = []
        #     for n_neighbors in n_neighbors_list:
        #         novelty_settings.append({"n_neighbors": n_neighbors})

        # f1score_best = 0
        # settings_best = None
        # reference_group_best = None
        # for settings in novelty_settings:
        #     # initialize reference group with the given start values
        #     reference_group = copy.deepcopy(y)
        #     # novelty detection
        #     settings_final = {
        #         "metric": "braycurtis",
        #         "n_neighbors": 2,
        #         **settings,
        #     }

        #     X_train = X[reference_group == True]
        #     X_test = X[reference_group == False]

        #     # find outliers (ones that shall not be in the reference)
        #     lof = LocalOutlierFactor(novelty=True, **settings_final)
        #     lof.fit(X_train)
        #     y_test = lof.predict(X_test)

        #     # modify reference
        #     reference_group[reference_group == False] = y_test == 1

        #     # calculate new f1score
        #     results = two_groups_differentiation(
        #         X, copy.deepcopy(reference_group), groups
        #     )
        #     f1score = results["f1score"]
        #     print(f"Novelty settings: {settings_final}")
        #     print(f"Novelty f1score: {f1score}")
        #     print(pd.Series(reference_group).value_counts())
        #     if not all(reference_group):
        #         if f1score > f1score_best:
        #             f1score_best = f1score
        #             settings_best = settings
        #             reference_group_best = copy.deepcopy(reference_group)
        # print(f"BEST Novelty settings: {settings_best}")
        # print(f"BEST Novelty f1score: {f1score_best}")
        # return reference_group_best
        
        # initialize reference group with the given start values
        reference_group = copy.deepcopy(y)
        # novelty detection
        if settings is None:
            settings = {}
        settings_final = {
            "metric": "braycurtis",
            "n_neighbors": 2,
            **settings,
        }

        X_train = X[reference_group == True]
        X_test = X[reference_group == False]

        # find outliers (ones that shall not be in the reference)
        lof = LocalOutlierFactor(novelty=True, **settings_final)
        lof.fit(X_train)
        y_test = lof.predict(X_test)

        # modify reference
        reference_group[reference_group == False] = y_test == 1

        return reference_group

    @property
    def differentiation_score(self):
        return self._differentiation_score

    @property
    def feature_columns(self):
        if self.log_ratio_bacteria is not None:
            feature_columns = self._feature_columns[
                self._feature_columns != self.log_ratio_bacteria
            ]
        else:
            feature_columns = self._feature_columns
        return feature_columns

    @feature_columns.setter
    def feature_columns(self, val):
        if isinstance(val, list):
            self._feature_columns = val
        elif isinstance(val, FeatureColumnsType):
            if val == FeatureColumnsType.BACTERIA:
                self._feature_columns = self.bacteria_columns
            elif val == FeatureColumnsType.METADATA:
                self._feature_columns = self.metadata_columns
            elif val == FeatureColumnsType.BACTERIA_AND_METADATA:
                self._feature_columns = self.bacteria_and_metadata_columns
        # but also update novelty detection result if it was used
        # self.reference_group_choice = self._reference_group_choice

    @property
    def bacteria_columns(self):
        return self.df.columns[self.df.columns.str.startswith("bacteria_")].to_numpy()

    @property
    def metadata_columns(self):
        return self.df.columns[self.df.columns.str.startswith("meta_")].to_numpy()

    @property
    def id_columns(self):
        return self.df.columns[self.df.columns.str.startswith("id_")].to_numpy()

    @property
    def other_columns(self):
        return self.df.columns[
            (~self.df.columns.str.startswith("bacteria_"))
            & (~self.df.columns.str.startswith("meta_"))
            & (~self.df.columns.str.startswith("id_"))
        ].to_numpy()

    @property
    def bacteria_and_metadata_columns(self):
        return np.concatenate((self.bacteria_columns, self.metadata_columns))

    @property
    def log_ratio_bacteria(self):
        return self._log_ratio_bacteria

    @log_ratio_bacteria.setter
    def log_ratio_bacteria(self, val):
        bacteria_columns_all = self.__df.columns[
            self.__df.columns.str.startswith("bacteria_")
        ].to_numpy()
        self._log_ratio_bacteria = val
        features = self.__df[bacteria_columns_all].values
        if val is not None:
            self.normalized = Normalization.NON_NORMALIZED

            for i, c in enumerate(bacteria_columns_all):
                self.df.loc[:, c] = features[:, i]
                if c != val:
                    self.df[c] = self.df.apply(
                        lambda row: math.log2(row[c] / row[val]), axis=1
                    )

            # remove reference, since these are abundances
            self.df = self.df.drop(columns=val, axis=1)

        else:

            for i, feature_column in enumerate(bacteria_columns_all):
                self.df.loc[:, feature_column] = features[:, i]
            # self.df = self.__df.copy(deep=True)

    @property
    def normalized(self):
        return self._normalized

    @normalized.setter
    def normalized(self, val):
        """Normalize all bacteria and metadata columns to 0-1 range.

        Normalization is performed both for novelty and trajectory features.

        Parameters
        ----------
        val : Normalization
            Value indicates whether to normalize or not (NORMALIZED or NON_NORMALIZED).
        """
        self._normalized = val
        features = self.__df[self.bacteria_and_metadata_columns].values

        if val == Normalization.NORMALIZED:
            from sklearn.preprocessing import normalize

            features = normalize(features, axis=0)

        for i, feature_column in enumerate(self.bacteria_and_metadata_columns):
            self.df.loc[:, feature_column] = features[:, i]

    def set_log_ratio_bacteria_with_least_crossings(self):
        def _crossings(x, y1, y2, degree=5):
            """
            To calculate crossing points between two abundance curves,
            we do the following:
            - approximate bacteria abundance over time with a curve,
            - get the indices of coordinates where two curves intersect.
            """
            f1 = np.poly1d(np.polyfit(x, y1, degree))
            f2 = np.poly1d(np.polyfit(x, y2, degree))

            x_new = np.linspace(min(x), max(x))
            # approximated bacteria abundances y1_new, y2_new
            y1_new = f1(x_new)
            y2_new = f2(x_new)

            # crossings
            idx = np.argwhere(np.diff(np.sign(y1_new - y2_new))).flatten()

            return idx

        from collections import defaultdict
        from itertools import product

        # abundance approximations
        x = self.__df.age_at_collection.tolist()

        number_of_crossings = defaultdict(lambda: 0)
        for b1, b2 in product(self.bacteria_columns, self.bacteria_columns):
            if b1 != b2:
                # bacteria abundances y1, y2
                y1 = self.__df[b1].tolist()
                y2 = self.__df[b2].tolist()
                number_of_crossings[b1] += len(_crossings(x, y1, y2, degree=5))

        # bacteria with the least number of crossings
        self.log_ratio_bacteria = min(number_of_crossings, key=number_of_crossings.get)

    def write_data(self, filename):
        with open(filename, "wb") as f:
            pickle.dump(self, f)

    @classmethod
    def read_data(filename):
        if os.path.exists(filename):
            with open(filename, "rb") as f:
                df = pickle.loads(f.read())
        else:
            df = None

        return df

    def z_score(self, column_name, sample_size=30):
        # TODO: above and below zscore values
        data = self.df[[column_name, "age_at_collection"]]
        data.age_at_collection //= sample_size
        from scipy.stats import zmap

        column_zscore = data.apply(
            lambda x: zmap(
                x[column_name],
                data[data.age_at_collection == x.age_at_collection][
                    column_name
                ].tolist(),
            )[0],
            axis=1,
        ).tolist()
        return column_zscore

    def plot_bacteria_abundances(
        self,
        number_of_columns=3,
        layout_settings=None,
        xaxis_settings=None,
        yaxis_settings=None,
    ):
        number_of_rows = len(self.bacteria_columns) // number_of_columns + 1

        layout_settings_default = dict(
            height=number_of_rows * 200,
            width=1500,
            plot_bgcolor="rgba(255,255,255,255)",
            paper_bgcolor="rgba(255,255,255,255)",
            margin=dict(l=0, r=0, b=0, pad=0),
            title_text="Bacteria Abundances in the Dataset",
            font=dict(size=10),
            yaxis=dict(position=0.0),
        )
        if layout_settings is None:
            layout_settings = {}
        layout_settings_final = {**layout_settings_default, **layout_settings}
        if xaxis_settings is None:
            xaxis_settings = {}
        if yaxis_settings is None:
            yaxis_settings = {}
        xaxis_settings_final = {**self.axis_settings_default, **xaxis_settings}
        yaxis_settings_final = {**self.axis_settings_default, **yaxis_settings}

        fig = make_subplots(
            rows=number_of_rows,
            cols=number_of_columns,
            horizontal_spacing=0.1,
            subplot_titles=[
                "<br>".join(wrap(str(b), 50)) for b in list(self.bacteria_columns)
            ],
        )
        for idx, bacteria_name in enumerate(self.bacteria_columns):
            df = self.df[["age_at_collection", bacteria_name]]
            fig.add_trace(
                go.Scatter(
                    x=df.groupby(by="age_at_collection")
                    .agg(np.mean)[bacteria_name]
                    .index,
                    y=df.groupby(by="age_at_collection")
                    .agg(np.mean)[bacteria_name]
                    .values,
                    error_y=dict(
                        type="data",  # value of error bar given in data coordinates
                        array=df.groupby(by="age_at_collection")
                        .agg(np.std)[bacteria_name]
                        .values,
                        visible=True,
                    ),
                    name=self.nice_name(bacteria_name),
                    hovertemplate="Abundance: %{y:.2f}"
                    + f"<br>{self.time_unit.name}:"
                    + " %{x}",
                ),
                row=idx // number_of_columns + 1,
                col=idx % number_of_columns + 1,
            )
            fig.update_xaxes(
                title=f"Age at collection [{self.time_unit.name}]",
                row=idx // number_of_columns + 1,
                col=idx % number_of_columns + 1,
                **xaxis_settings_final,
            )
            fig.update_yaxes(
                title="Abundance value",
                row=idx // number_of_columns + 1,
                col=idx % number_of_columns + 1,
                **yaxis_settings_final,
            )

        for i in fig["layout"]["annotations"]:
            i["font"] = dict(size=10)

        fig.update_layout(**layout_settings_final)

        config = {
            "toImageButtonOptions": {
                "format": "svg",  # one of png, svg, jpeg, webp
                "filename": "bacteria_abundances",
                "height": layout_settings_final["height"],
                "width": layout_settings_final["width"],
                "scale": 1,  # Multiply title/legend/axis/canvas sizes by this factor
            }
        }
        results = {"fig": fig, "config": config}

        return results

    def plot_bacteria_abundance_heatmaps(
        self,
        relative_values=False,
        fillna=False,
        dropna=False,
        avg_fn=np.median,
        layout_settings=None,
    ):
        layout_settings_default = dict(
            height=20 * len(self.bacteria_columns),
            width=1500,
        )
        if layout_settings is None:
            layout_settings = {}
        layout_settings_final = {**layout_settings_default, **layout_settings}

        # extract just the important columns for the heatmap
        df = self.df[list(self.bacteria_columns) + ["subjectID", "age_at_collection"]]
        # replace long bacteria names with nice names
        # df = df.rename({b: self.nice_name(b) for b in self.bacteria_columns}, axis=1)
        # update the bacteria_names with short names

        def fill_collected(row):
            """Ïf we use bigger time units, then we need to find a median when collapsing all the samples in that time interval into one box in the heatmap"""
            val = df[df["age_at_collection"] == row["age_at_collection"]][
                row["bacteria_name"]
            ].values
            row["bacteria_value"] = avg_fn(val) if len(val) != 0 else np.nan
            return row

        x, y = np.meshgrid(
            self.bacteria_columns, range(int(max(df.age_at_collection)) + 1)
        )

        df_heatmap = pd.DataFrame(
            data={
                "bacteria_name": x.flatten(),
                "age_at_collection": y.flatten(),
                "bacteria_value": np.nan,
            }
        )
        df_heatmap = df_heatmap.sort_values(by=["bacteria_name", "age_at_collection"])
        # if dropna:
        #     df_heatmap = df_heatmap.dropna()
        df_heatmap = df_heatmap.fillna(0)
        df_heatmap = df_heatmap.apply(lambda row: fill_collected(row), axis=1)

        # create new column bacteria_name_cat in order to sort dataframe by bacteria importance
        df_heatmap["bacteria_name_cat"] = pd.Categorical(
            df_heatmap["bacteria_name"],
            categories=self.bacteria_columns,  # order of bacteria is imposed by the list
            ordered=True,
        )
        df_heatmap = df_heatmap.sort_values("bacteria_name_cat")
        df_heatmap = df_heatmap[df_heatmap.age_at_collection > 0]

        if relative_values:
            X = df_heatmap.bacteria_value.values
            X = X.reshape(len(self.bacteria_columns), -1)
            xmin = np.nanmin(X, axis=1).reshape(len(self.bacteria_columns), 1)
            xmax = np.nanmax(X, axis=1).reshape(len(self.bacteria_columns), 1)
            X_std = (X - xmin) / (xmax - xmin + 1e-10)
            df_heatmap.bacteria_value = X_std.flatten()

        # plot top absolute
        df_heatmap = df_heatmap[
            ["age_at_collection", "bacteria_name_cat", "bacteria_value"]
        ]
        bacteria_names = list(map(self.nice_name, self.bacteria_columns))
        uniquify(bacteria_names, suffs=(f"_{x!s}" for x in range(1, 100)))
        bacteria_names = np.array(bacteria_names)
        bacteria_names_mapping = {
            k: v for k, v in zip(self.bacteria_columns, bacteria_names)
        }

        def _rename_bacteria(row):
            row["bacteria_name_cat"] = bacteria_names_mapping[row["bacteria_name_cat"]]
            return row

        df_heatmap = df_heatmap.apply(lambda row: _rename_bacteria(row), axis=1)

        df_heatmap_pivot = df_heatmap.pivot(
            "bacteria_name_cat", "age_at_collection", "bacteria_value"
        )

        if fillna:
            df_heatmap_pivot = df_heatmap_pivot.fillna(0)

        if dropna:
            df_heatmap_pivot = df_heatmap_pivot.dropna(axis=1, how="all")

        xaxis_mapping = {v: k for k, v in enumerate(df_heatmap_pivot.columns.tolist())}
        fig = px.imshow(
            df_heatmap_pivot.values,
            labels=dict(
                x=f"Age [{self.time_unit.name}]", y="Bacteria Name", color="Abundance"
            ),
            x=[xaxis_mapping[i] for i in df_heatmap_pivot.columns.tolist()],
            y=df_heatmap_pivot.index.tolist(),
            title="Abundances",
            **layout_settings_final,
        )
        if dropna:
            fig.update_xaxes(
                side="bottom",
                tickvals=[xaxis_mapping[i] for i in df_heatmap_pivot.columns.tolist()],
                ticktext=df_heatmap_pivot.columns.tolist(),
            )

        config = {
            "toImageButtonOptions": {
                "format": "svg",  # one of png, svg, jpeg, webp
                "filename": "bacteria_abundances_heatmap",
                "height": layout_settings_final["height"],
                "width": layout_settings_final["width"],
                "scale": 1,  # Multiply title/legend/axis/canvas sizes by this factor
            }
        }
        results = {"fig": fig, "config": config}

        return results

    def plot_ultradense_longitudinal_data(
        self,
        number_of_columns=6,
        number_of_bacteria=20,
        color_palette="tab20",
        layout_settings=None,
        xaxis_settings=None,
        yaxis_settings=None,
    ):
        number_of_rows = len(self.df.subjectID.unique()) // number_of_columns + int(
            len(self.df.subjectID.unique()) % number_of_columns > 0
        )

        # plotly settings
        layout_settings_default = dict(
            height=350 * number_of_rows,
            width=1500,
            plot_bgcolor="rgba(0,0,0,0)",
            title_text="Ultradense Longitudinal Data",
        )
        if layout_settings is None:
            layout_settings = {}
        layout_settings_final = {**layout_settings_default, **layout_settings}
        if xaxis_settings is None:
            xaxis_settings = {}
        if yaxis_settings is None:
            yaxis_settings = {}
        xaxis_settings_final = {**self.axis_settings_default, **xaxis_settings}
        yaxis_settings_final = {**self.axis_settings_default, **yaxis_settings}

        # limit to plot 20 bacteria
        bacteria_names = self.bacteria_columns[:number_of_bacteria]

        cmap = plt.cm.get_cmap(color_palette, len(bacteria_names))
        colors_dict = dict([(b, cmap(i)) for i, b in enumerate(bacteria_names)])

        fig = make_subplots(
            rows=number_of_rows,
            cols=number_of_columns,
            shared_xaxes=True,
            shared_yaxes=True,
            vertical_spacing=0.08,
            horizontal_spacing=0.01,
        )

        for idx, infant in enumerate(self.df.subjectID.unique()):
            i, j = (
                idx // number_of_columns,
                idx % number_of_columns,
            )

            df1 = self.df.reset_index()
            df1 = df1[df1.subjectID == infant].sort_values("age_at_collection")

            if len(df1) == 1:

                for b in bacteria_names:

                    fig.add_trace(
                        go.Scatter(
                            x=list(df1.age_at_collection.values),
                            y=list(df1[b].values) * 2,
                            text=list(map(lambda x: self.nice_name(x), bacteria_names)),
                            mode="lines",
                            marker_color=f"rgba{colors_dict[b]}",
                            name=self.nice_name(b),
                            legendgroup=self.nice_name(b),
                            showlegend=True if idx == 0 else False,
                            stackgroup="one",  # define stack group
                            hovertemplate=f"Taxa: {self.nice_name(b)}"
                            + "<br>Abundance: %{y:.2f}"
                            + f"<br>{self.time_unit.name}:"
                            + " %{x}",
                        ),
                        row=i + 1,
                        col=j + 1,
                    )

                    fig.update_xaxes(
                        title=infant,
                        row=i + 1,
                        col=j + 1,
                        **xaxis_settings_final,
                    )
                    fig.update_yaxes(
                        **yaxis_settings_final,
                    )

            else:
                for b in bacteria_names:
                    fig.add_trace(
                        go.Scatter(
                            x=list(df1.age_at_collection.values),
                            y=list(df1[b].values),
                            text=list(map(lambda x: self.nice_name(x), bacteria_names)),
                            mode="lines",
                            marker_color=f"rgba{colors_dict[b]}",
                            name=self.nice_name(b),
                            legendgroup=self.nice_name(b),
                            showlegend=True if idx == 0 else False,
                            stackgroup="one",
                            hovertemplate=f"Taxa: {self.nice_name(b)}"
                            + "<br>Abundance: %{y:.2f}"
                            + f"<br>{self.time_unit.name}:"
                            + " %{x}",
                        ),
                        row=i + 1,
                        col=j + 1,
                    )

                    fig.update_xaxes(
                        title=infant,
                        row=i + 1,
                        col=j + 1,
                        **xaxis_settings_final,
                    )
                    fig.update_yaxes(
                        **yaxis_settings_final,
                    )

        fig.update_layout(**layout_settings_final)

        for i in fig["layout"]["annotations"]:
            i["font"] = dict(size=8, color="#000000")

        config = {
            "toImageButtonOptions": {
                "format": "svg",  # one of png, svg, jpeg, webp
                "filename": "bacteria_abundances_longitudinal",
                "height": layout_settings_final["height"],
                "width": layout_settings_final["width"],
                "scale": 1,  # Multiply title/legend/axis/canvas sizes by this factor
            }
        }
        results = {"fig": fig, "config": config}

        return results

    def embedding_to_latent_space(
        self,
        color_column_name=None,
        embedding_model=PCA(n_components=3),
        embedding_dimension=2,
        layout_settings=None,
        xaxis_settings=None,
        yaxis_settings=None,
    ):
        fig = go.Figure()

        color_column_name = color_column_name or "group"
        subjectIDs = np.array(self.df.subjectID.tolist())
        sampleIDs = np.array(self.df.sampleID.tolist())
        X = self.df[self.feature_columns].values
        X_emb = embedding_model.fit_transform(X)

        if isinstance(embedding_model, PCA):
            xaxis_label = f"PC1 - {embedding_model.explained_variance_ratio_[0]*100:.1f}% explained variance"
            yaxis_label = f"PC2 - {embedding_model.explained_variance_ratio_[1]*100:.1f}% explained variance"
        else:
            xaxis_label = "1st dimension"
            yaxis_label = "2nd dimension"

        # plotly settings
        layout_settings_default = dict(
            height=600,
            width=600,
            barmode="stack",
            uniformtext=dict(mode="hide", minsize=10),
            plot_bgcolor="rgba(0,0,0,0)",
            paper_bgcolor="rgba(0,0,0,0)",
            margin=dict(l=0, r=0, b=0, pad=0),
            title_text=f"Embedding in {embedding_dimension}D space",
            annotations=[
                dict(
                    x=0.5,
                    y=1.05,
                    align="right",
                    valign="top",
                    text="colored by " + color_column_name,
                    showarrow=False,
                    xref="paper",
                    yref="paper",
                    xanchor="center",
                    yanchor="top",
                )
            ],
        )
        if layout_settings is None:
            layout_settings = {}
        layout_settings_final = {**layout_settings_default, **layout_settings}
        if xaxis_settings is None:
            xaxis_settings = {}
        if yaxis_settings is None:
            yaxis_settings = {}
        xaxis_settings_final = {**self.axis_settings_default, **xaxis_settings}
        yaxis_settings_final = {**self.axis_settings_default, **yaxis_settings}

        if embedding_dimension == 2:
            if color_column_name:
                for g in self.df[color_column_name].unique():
                    idx = np.where(np.array(self.df[color_column_name].values) == g)[0]
                    fig.add_trace(
                        go.Scatter(
                            x=X_emb[:, 0][idx],
                            y=X_emb[:, 1][idx],
                            name=str(g) or "NaN",
                            mode="markers",
                            text=[
                                f"<b>SampleID</b> {i}<br><b>SubjectID</b> {j}<br>"
                                for i, j in zip(sampleIDs[idx], subjectIDs[idx])
                            ],
                            hovertemplate="%{text}"
                            + f"<b>Group ({color_column_name}): {g}</b><br>"
                            + "<b>x</b>: %{x:.2f}<br>"
                            + "<b>y</b>: %{y:.2f}<br>",
                        )
                    )
            else:
                fig.add_trace(
                    go.Scatter(
                        x=X_emb[:, 0],
                        y=X_emb[:, 1],
                        # name=sampleIDs,
                        mode="markers",
                        text=[
                            f"<b>SampleID</b> {i}<br><b>SubjectID</b> {j}<br>"
                            for i, j in zip(sampleIDs, subjectIDs)
                        ],
                        hovertemplate="%{text}"
                        + "<b>x</b>: %{x:.2f}<br>"
                        + "<b>y</b>: %{y:.2f}<br>",
                    )
                )
        elif embedding_dimension == 3:
            if color_column_name:
                for g in self.df[color_column_name].unique():
                    idx = np.where(np.array(self.df[color_column_name].values) == g)[0]

                    fig.add_trace(
                        go.Scatter3d(
                            x=X_emb[:, 0][idx],
                            y=X_emb[:, 1][idx],
                            z=X_emb[:, 2][idx],
                            name=str(g) or "NaN",
                            mode="markers",
                            marker=dict(size=5, opacity=0.8),
                            text=[
                                f"<b>SampleID</b> {i}<br><b>SubjectID</b> {j}<br>"
                                for i, j in zip(sampleIDs[idx], subjectIDs[idx])
                            ],
                            hovertemplate="%{text}"
                            + f"<b>Group ({color_column_name}): {g}</b><br>"
                            + "<b>x</b>: %{x:.2f}<br>"
                            + "<b>y</b>: %{y:.2f}<br>",
                        )
                    )
            else:
                fig.add_trace(
                    go.Scatter3d(
                        x=X_emb[:, 0],
                        y=X_emb[:, 1],
                        z=X_emb[:, 2],
                        mode="markers",
                        marker=dict(size=5, opacity=0.9),
                        text=[
                            f"<b>SampleID</b> {i}<br><b>SubjectID</b> {j}<br>"
                            for i, j in zip(sampleIDs, subjectIDs)
                        ],
                        hovertemplate="%{text}"
                        + "<b>x</b>: %{x:.2f}<br>"
                        + "<b>y</b>: %{y:.2f}<br>",
                    )
                )
        else:
            raise NotImplemented(
                f"Dimension {embedding_dimension} not supported for visualization :)"
            )

        fig.update_xaxes(
            title=xaxis_label,
            **xaxis_settings_final,
        )
        fig.update_yaxes(
            title=yaxis_label,
            **yaxis_settings_final,
        )

        fig.update_layout(**layout_settings_final)

        config = {
            "toImageButtonOptions": {
                "format": "svg",  # one of png, svg, jpeg, webp
                "filename": "embedding_to_latent_space",
                "height": layout_settings_final["height"],
                "width": layout_settings_final["width"],
                "scale": 1,  # Multiply title/legend/axis/canvas sizes by this factor
            }
        }

        results = {"fig": fig, "config": config}

        return results

    def embeddings_interactive_selection_notebook(
        self,
        embedding_model=PCA(n_components=2),
        layout_settings=None,
        xaxis_settings=None,
        yaxis_settings=None,
    ):
        fig = go.Figure()

        subjectIDs = np.array(self.df.subjectID.tolist())
        sampleIDs = np.array(self.df.sampleID.tolist())
        X = self.df[self.feature_columns].values
        X_emb = embedding_model.fit_transform(X)

        if isinstance(embedding_model, PCA):
            xaxis_label = f"PC1 - {embedding_model.explained_variance_ratio_[0]*100:.1f}% explained variance"
            yaxis_label = f"PC2 - {embedding_model.explained_variance_ratio_[1]*100:.1f}% explained variance"
        else:
            xaxis_label = "1st dimension"
            yaxis_label = "2nd dimension"

        layout_settings_default = dict(
            height=600,
            width=900,
            barmode="stack",
            uniformtext=dict(mode="hide", minsize=10),
            plot_bgcolor="rgba(0,0,0,0)",
            paper_bgcolor="rgba(0,0,0,0)",
            margin=dict(l=0, r=0, b=0, pad=0),
            title_text="Embedding in 2D space",
        )
        if layout_settings is None:
            layout_settings = {}
        layout_settings_final = {**layout_settings_default, **layout_settings}
        if xaxis_settings is None:
            xaxis_settings = {}
        if yaxis_settings is None:
            yaxis_settings = {}
        xaxis_settings_final = {**self.axis_settings_default, **xaxis_settings}
        yaxis_settings_final = {**self.axis_settings_default, **yaxis_settings}

        fig.add_trace(
            go.Scatter(
                x=X_emb[:, 0],
                y=X_emb[:, 1],
                # name=sampleIDs,
                mode="markers",
                text=[
                    f"<b>SampleID</b> {i}<br><b>SubjectID</b> {j}<br>"
                    for i, j in zip(sampleIDs, subjectIDs)
                ],
                hovertemplate="%{text}"
                + "<b>x</b>: %{x:.2f}<br>"
                + "<b>y</b>: %{y:.2f}<br>",
            )
        )

        fig.update_xaxes(
            title=xaxis_label,
            **xaxis_settings_final,
        )
        fig.update_yaxes(
            title=yaxis_label,
            **yaxis_settings_final,
        )

        fig.update_layout(**layout_settings_final)

        f = go.FigureWidget(fig)

        # Create a table FigureWidget that updates on selection from points in the scatter plot of f
        t = go.FigureWidget(
            [
                go.Table(
                    header=dict(
                        values=["sampleID", "subjectID"],
                        fill=dict(color="#C2D4FF"),
                        align=["left"] * 5,
                    ),
                    cells=dict(
                        values=[self.df[col] for col in ["sampleID", "subjectID"]],
                        fill=dict(color="#F5F8FF"),
                        align=["left"] * 5,
                    ),
                )
            ]
        )

        def selection_fn(trace, points, selector):
            results = self.selection_embeddings(t)(trace, points, selector)

            # plot the result of reference analysis with feature_columns_for_reference
            results["fig"].show(config=results["config"])
            results["img_src"].show()

        f.data[0].on_selection(selection_fn)

        if "selected" in self.df.columns:
            self.df.drop(columns="selected", inplace=True)

        plt.close("all")

        config = {
            "toImageButtonOptions": {
                "format": "svg",  # one of png, svg, jpeg, webp
                "filename": "embedding_to_latent_space",
                "height": layout_settings_final["height"],
                "width": layout_settings_final["width"],
                "scale": 1,  # Multiply title/legend/axis/canvas sizes by this factor
            }
        }

        results = {"vbox": VBox((f, t)), "config": config}

        # Put everything together
        return results

    def selection_embeddings(self, t):
        def _inner(trace, points, selector):
            if hasattr(points, "point_inds"):
                t.data[0].cells.values = [
                    self.df.loc[points.point_inds][col]
                    for col in ["sampleID", "subjectID"]
                ]
            else:
                t.data[0].cells.values = [
                    self.df.loc[points][col] for col in ["sampleID", "subjectID"]
                ]

            df_selected = pd.DataFrame(
                data={
                    "sampleID": t.data[0].cells.values[0],
                    "subjectID": t.data[0].cells.values[1],
                }
            )

            plt.clf()
            # create new column called selected to use for reference analysis: True - selected, False - not selected
            self.df["selected"] = False
            self.df.loc[
                self.df["sampleID"].isin(df_selected["sampleID"]), "selected"
            ] = True
            groups = self.df.subjectID.values

            results = two_groups_differentiation(
                self.df[self.feature_columns],
                self.df["selected"],
                groups,
                nice_name=self.nice_name,
                plot=True,
            )

            return results

        return _inner


def two_groups_differentiation(
    X,
    y,
    groups,
    y_column=None,
    nice_name=lambda x: x,
    settings=None,
    plot=False,
    layout_settings=None,
    xaxis_settings=None,
    yaxis_settings=None,
):
    if isinstance(X, pd.DataFrame):
        x_columns = X.columns
        X = X.values
    else:
        x_columns = [f"feature_{i}" for i in range(X.shape[1])]

    if isinstance(y, pd.Series):
        y_column = y_column or y.name
        y = y.values.astype(int)
    else:
        y_column = "main group"
        y = y.astype(int)

    if isinstance(groups, pd.Series):
        groups = groups.values

    if settings is None:
        settings = {}
    settings_final = {
        "n_estimators": 140,
        "max_samples": 0.8,
        "random_state": RANDOM_STATE,
        **settings,
    }

    rfc = RandomForestClassifier(**settings_final)
    splits = GroupShuffleSplit(random_state=RANDOM_STATE)
    f1_scores = cross_val_score(
        rfc,
        X,
        y,
        scoring="f1",
        groups=groups,
        cv=splits,
    )
    accuracy_scores = cross_val_score(
        rfc,
        X,
        y,
        scoring="accuracy",
        groups=groups,
        cv=splits,
    )

    rfc.fit(X, y)

    fig, config, img_src = None, None, None
    if plot:
        max_display = 20
        layout_settings_default = dict(
            height=1000,
            width=800,
            plot_bgcolor="rgba(0,0,0,0)",
            margin=dict(l=0, r=0, b=0, pad=0),
            title_text="Classification Important Features",
        )
        axis_settings_default = dict(
            tick0=0,
            mirror=True,
            # dtick=2,
            showline=True,
            linecolor="lightgrey",
            gridcolor="lightgrey",
            zeroline=True,
            zerolinecolor="lightgrey",
            showspikes=True,
            spikecolor="gray",
            autorange=True,
        )
        if layout_settings is None:
            layout_settings = {}
        layout_settings_final = {**layout_settings_default, **layout_settings}
        if xaxis_settings is None:
            xaxis_settings = {}
        if yaxis_settings is None:
            yaxis_settings = {}
        xaxis_settings_final = {**axis_settings_default, **xaxis_settings}
        yaxis_settings_final = {**axis_settings_default, **yaxis_settings}

        shap.initjs()

        explainer = shap.TreeExplainer(rfc)
        shap_values = explainer.shap_values(X)
        fig, ax = plt.subplots()

        shap.summary_plot(
            shap_values[0],
            features=X,
            class_names=[y_column, f"non-{y_column}"],
            show=False,
            max_display=max_display,
        )

        max_limit = max(20, min(max_display, len(x_columns)))

        fig = mpl_to_plotly(fig)

        fig.update_xaxes(
            title="SHAP value (impact on model output)",
            **xaxis_settings_final,
        )

        fig.update_yaxes(
            title="Features",
            tickmode="array",
            tickvals=list(range(0, max_limit)),
            ticktext=list(
                map(
                    nice_name,
                    list(
                        np.array(x_columns)[np.argsort(np.abs(shap_values[0]).mean(0))]
                    )[::-1][:max_limit][::-1],
                )
            ),
            **yaxis_settings_final,
        )
        fig.update_layout(**layout_settings_final)

        f1_scores = cross_val_score(
            rfc,
            X,
            y,
            scoring="f1",
            groups=groups,
            # cv=splits,
        )
        accuracy_scores = cross_val_score(
            rfc,
            X,
            y,
            scoring="accuracy",
            groups=groups,
            # cv=splits,
        )
        y_pred = cross_val_predict(
            rfc,
            X,
            y,
            groups=groups,
            # cv=splits,
        )
        
        cm = confusion_matrix(y_pred, y)

        img_src = plot_confusion_matrix(
            cm, [y_column, f"non-{y_column}"], "Confusion matrix"
        )

        config = {
            "toImageButtonOptions": {
                "format": "svg",  # one of png, svg, jpeg, webp
                "filename": "embedding_interactive",
                "height": layout_settings_final["height"],
                "width": layout_settings_final["width"],
                "scale": 1,  # Multiply title/legend/axis/canvas sizes by this factor
            }
        }

    results = {
        "accuracy": np.mean(accuracy_scores),
        "f1score": np.mean(f1_scores),
        "fig": fig,
        "config": config,
        "img_src": img_src,
    }
    return results


def plot_confusion_matrix(cm, classes, title):
    # cm : confusion matrix list(list)
    # classes : name of the data list(str)
    # title : title for the heatmap

    data = go.Heatmap(
        z=cm,
        y=classes,
        x=classes,
        colorscale=["#ffffff", "#F52757"],
    )

    annotations = []

    cm = np.divide(cm.astype("float"), cm.sum())#[:, np.newaxis]
    fmt = ".1%"
    
    thresh = cm.max() / 2.0

    for i, j in itertools.product(range(cm.shape[0]), range(cm.shape[1])):
        annotations.append(
            {
                "x": classes[j],
                "y": classes[i],
                "font": {
                    "color": "white" if cm[i, j] > thresh else "black",
                    "size": 20,
                },
                "text": format(cm[i, j], fmt),
                "xref": "x1",
                "yref": "y1",
                "showarrow": False,
            }
        )

    layout = {
        "title": title,
        "xaxis": {"title": "Predicted value"},
        "yaxis": {"title": "Real value"},
        "annotations": annotations,
        "height": 500,
        "width": 500,
    }
    fig = go.Figure(data=data, layout=layout)
    return fig


def uniquify(seq, suffs=None):
    """Make all the items unique by adding a suffix (1, 2, etc).

    `seq` is mutable sequence of strings.
    `suffs` is an optional alternative suffix iterable.
    """
    not_unique = [k for k, v in Counter(seq).items() if v > 1]
    # suffix generator dict - e.g., {'name': <my_gen>, 'zip': <my_gen>}
    suff_gens = dict(zip(not_unique, tee(suffs, len(not_unique))))
    for idx, s in enumerate(seq):
        try:
            suffix = str(next(suff_gens[s]))
        except KeyError:
            # s was unique
            continue
        else:
            seq[idx] += suffix
